#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 13:37:00 2020

@author: 
Dr. Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
GitHub: https://github.com/MNGuenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import os
import ellc
from astropy.constants import G
from astropy import units as u

#::: my modules
from ..exoworlds_rdx.lightcurves import lightcurve_tools as lct
from . import translator, plotter, defaults
from .. import computer

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




'''
A collection of generative models (lightcurve / RV / flare) for ease-of-use
'''


###############################################################################
#::: defaults from kwargs
###############################################################################
def get_defaults(time, params, settings, options, **kwargs):    
    '''
    Input:
    ------
    params : dict
        with the following keys
    settings : dict
        with the following keys
        
    params keys or kwargs
    ---------------------    
    rr : float
        radius ratio
    rsuma : float
        sum of raddi over a
    cosi : float
        cosine of orbital inclination
    epoch : float
        epoch in days
    period : float
        period in days
    K : float
        RV semi-amplitude in km/s
    f_c : float
        sqrt(e) cos(omega)
    f_s : float
        sqrt(e) sin(omega)
    dil : float
        dilution, D_0 = 1 - (Ftarget / (Ftarget + Fblend))
        default is 0
    sbratio : float
        surface brightness ratio
        default is 0
    ldc : float or list
        limb darkening coefficients
        default is [0.4804, 0.1867]
        
    settings keys or kwargs
    ---------------------    
    ld : str
        limb darkening law
        default is 'quad'
        
    options keys or kwargs
    ---------------------    
    show_plot : bool
        show the plot in the terminal, or close it
        default is False
    save_plot : bool
        save the plot to a file, or not
        default is False
    save_csv : bool
        save the lightcurve to a file, or not
        default is False
    fname_plot : str
        filename for the plot
        default is 'lightcurve.pdf'
    fname_csv : str
        filename for the csv
        default is 'lightcurve.csv'
    '''
    
    if options is None: options = {}
    if 'show_plot' not in options: options['show_plot'] = False
    if 'save_plot' not in options: options['save_plot'] = False
    if 'save_csv' not in options: options['save_csv'] = False
    if 'fname_plot' not in options: options['fname_plot'] = 'output.pdf'
    if 'fname_csv' not in options: options['fname_csv'] = 'output.csv'
    for key in kwargs: 
        if key in options:
            options[key] = kwargs[key]
    
    return params, settings, options



###############################################################################
#::: make an ellc model lightcurve
###############################################################################
def make_transit_model(time, inst, params, settings, **kwargs):
    '''
    Inputs:
    -------
    time : array of float
        time in days
        
    Optional Inputs:
    ----------------
    flux : array of float
        flux of the 'underlying' lightcurve
    flux_err : array of float
        flux error of the 'underlying' lightcurve
    see get_defaults()
        
    Returns:
    --------
    model_flux : array of float
        relative flux of the model
    '''
    
    def fct(time): 
        params_ellc = translator.translate_alles_to_ellc(params, settings)
        model_flux = computer.calculate_model(params_ellc, inst, key='flux', xx=time, settings=settings)
        return model_flux
    
    model_flux = fct(time)
    flux2 = flux + model_flux - 1 
        
    if options['show_plot'] or options['save_plot']:   

        #::: get model on fine grid
        time_grid = np.linspace(time[0], time[-1], 10001)
        model_flux_grid = ellc_lc_short(time_grid)
        
        #::: get phase-folded data
        phase, phaseflux, phaseflux_err, N, phi = lct.phase_fold(time, flux2, params['period'], params['epoch'], dt=0.002, ferr_type='medsig', ferr_style='sem', sigmaclip=True)

        #::: get phase-folded model on fine grid from phase -0.25 to 0.75
        phase_grid = np.linspace(-0.25, 0.75, 10001 )
        model_phaseflux_grid = ellc_lc_short( params['epoch']+phase_grid*params['period'] ) #need to input the phase as time domain for ellc
                    
        #::: plot all
        fig = plt.figure(figsize=(15,10), tight_layout=True)
        gs = gridspec.GridSpec(3, 3)
        plotter.plot_lc_full(fig.add_subplot(gs[0,:]), time, flux2, flux_err, time_grid, model_flux_grid)
        plotter.plot_lc_phase(fig.add_subplot(gs[1,0]), phi, flux2, phase, phaseflux, phaseflux_err, phase_grid, model_phaseflux_grid)
        plotter.plot_lc_phasezoom(fig.add_subplot(gs[1,1]), phi, flux2, phase, phaseflux, phaseflux_err, phase_grid, model_phaseflux_grid, params['period']*24.*60.)
        plotter.plot_info(fig.add_subplot(gs[2,0]), text=0, params=params)
        plotter.plot_info(fig.add_subplot(gs[2,1]), text=1, params=params)
        
        if options['save_plot']:
            if len(os.path.dirname(options['fname_plot']))>0 and not os.path.exists(os.path.dirname(options['fname_plot'])): os.makedirs(os.path.dirname(options['fname_plot']))
            fig.savefig(options['fname_plot'])
            plt.close(fig)

        if options['show_plot']: plt.show(fig)
        else: plt.close(fig)
         
    if options['save_csv']:
        if len(os.path.dirname(options['fname_csv']))>0 and not os.path.exists(os.path.dirname(options['fname_csv'])): os.makedirs(os.path.dirname(options['fname_csv']))
        X = np.column_stack((time, flux2, flux_err))
        np.savetxt(options['fname_csv'], X, delimiter=',')


    return flux2

    
    
###############################################################################
#::: inject an ellc model lightcurve
###############################################################################
def inject_transit_model(time, flux, flux_err, 
                            params=None, settings=None, options=None, 
                            **kwargs):
    '''
    Wrapper around make_lc_model()
    '''
    
    return make_lightcurve_model(time, flux=flux, flux_err=flux_err, 
                                 params=None, settings=None, options=None, 
                                 **kwargs)
    


   
###############################################################################
#::: make an ellc model lightcurve
###############################################################################
def make_rv_model(time, 
                  rv=None, rv_err=None, 
                  epoch=0., period=1., 
                  R_companion=1., M_companion=1., 
                  R_companion_unit='Rearth', M_companion_unit='Mearth',
                  R_host=1., M_host=1., 
                  sbratio=0,
                  incl=90, 
                  ecc=0,
                  omega=0,
                  dil=0,
                  ldc=[0.4804, 0.1867],
                  ld='quad',
                  show_plot=False, save_plot=False, save_csv=False,
                  fname_plot='rv.pdf', fname_csv='rv.csv'):
    '''
    Inputs:
    -------
    time : array of float
        time in days
    rv : array of float
        RV of the 'underlying' series
    rv_err : array of float
        error of RV of the 'underlying' series
    epoch : float
        epoch in days
    period : float
        period in days
    R_companion: float
        radius of the planet, in R_earth
        
    Optional Inputs:
    ----------------
    M_companion: float
        mass of the planet, in M_earth
        default is 0
    R_host : float
        radius of the star, in R_sun
        default is 1
    M_host: float
        mass of the star, in M_sun
        default is 1
    show_plot : bool
        show the plot in the terminal, or close it
        default is False
    save_plot : bool
        save the plot to a file, or not
        default is False
    save_csv : bool
        save the lightcurve to a file, or not
        default is False
        
    Returns:
    --------
    flux2 : array of float
        relative flux with injected signal
    '''
    
    if rv is None: rv = np.zeros_like(time)
    if rv_err is None: rv_err = np.zeros_like(time)
        
    params = translate(quiet=True, R_companion=R_companion, M_companion=M_companion, R_companion_unit=R_companion_unit, M_companion_unit=M_companion_unit, R_host=R_host, M_host=M_host, epoch=epoch, period=period, incl=incl, ecc=ecc, omega=omega, ldc=ldc, ld=ld)   
    def ellc_rv_short(time): 
        return ellc.rv(t_obs=time, radius_1=params['R_host/a'], radius_2=params['R_companion/a'], sbratio=sbratio, incl=params['incl'], t_zero=epoch, period=params['period'], a=params['a'], q=1, f_c=params['f_c'], f_s=params['f_s'], ldc_1=ldc, ldc_2=None, gdc_1=None, gdc_2=None, didt=None, domdt=None, rotfac_1=1, rotfac_2=1, hf_1=1.5, hf_2=1.5, bfac_1=None, bfac_2=None, heat_1=None, heat_2=None, lambda_1=None, lambda_2=None, vsini_1=None, vsini_2=None, t_exp=None, n_int=None,  grid_1='default', grid_2='default', ld_1=ld, ld_2=None, shape_1='sphere', shape_2='sphere', spots_1=None, spots_2=None, verbose=1)[0]
    
    model_rv = ellc_rv_short(time)
    rv += model_rv
    
    if show_plot or save_plot:
        
        #::: get phase 
        phi = lct.calc_phase(time, period, epoch)
        phi[phi>0.75] -= 1.

        #::: get model on fine grid
        time_grid = np.linspace( time[0], time[-1], 10001 )
        model_rv_grid = ellc_rv_short(time_grid)
    
        #::: get model on fine grid from phase -0.25 to 0.75
        phase_grid = np.linspace(-0.25, 0.75, 10001 )
        model_phaserv_grid = ellc_rv_short(epoch+phase_grid*period)
    
        #::: plot
        fig = plt.figure(figsize=(15,10), tight_layout=True)
        gs = gridspec.GridSpec(3, 3)
        plotter.plot_rv_full(fig.add_subplot(gs[0,:]), time, rv, rv_err, time_grid, model_rv_grid)
        plotter.plot_rv_phase(fig.add_subplot(gs[1,0]), phi, rv, rv_err, phase_grid, model_phaserv_grid)
        plotter.plot_info(fig.add_subplot(gs[2,0]), text=0, params=params)
        plotter.plot_info(fig.add_subplot(gs[2,1]), text=1, params=params)
            
        if save_plot:
            if len(os.path.dirname(fname_plot))>0 and not os.path.exists(os.path.dirname(fname_plot)): os.makedirs(os.path.dirname(fname_plot))
            fig.savefig(fname_plot)

        if show_plot: plt.show(fig)
        else: plt.close(fig)
    
    if save_csv:
        if len(os.path.dirname(fname_csv))>0 and not os.path.exists(os.path.dirname(fname_csv)): os.makedirs(os.path.dirname(fname_csv))
        X = np.column_stack((time, rv, rv_err))
        np.savetxt(fname_csv, X, delimiter=',')


    return rv



###############################################################################
#::: inject an ellc model RV series
###############################################################################
def inject_rv_model(time, rv, rv_err, **kwargs):
    '''
    Wrapper around make_lc_model()
    '''
    
    return make_rv_model(time, rv = rv, rv_err = rv_err, **kwargs)
    


###############################################################################
#::: test
###############################################################################
# if __name__ == '__main__':
#     time = np.linspace(0,100,101)
#     make_lightcurve_model(time, period=30.7, R_host=0.4, M_host=0.4, show_plot=True)
    # make_rv_model(time, period=30.7, R_host=0.4, M_host=0.4, show_plot=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr  9 11:46:38 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt

from . import defaults, translator, generator
from .. import computer




class allesclass2():
    
    def __init__(self, datadir=None):
        if datadir is None:
            self.settings = {}
            self.params = {}
            self.params_host = {}
            self.data = {}
        else:
            pass #TODO
        
        
    #==========================================================================
    #::: fill gaps with defaults
    #==========================================================================
    def fill_settings(self):
        self.settings = defaults.fill_settings(self.settings)
        
    def fill_params(self):
        self.params = defaults.fill_params(self.params, self.settings)
        
    def fill_params_host(self):
        self.params_host = defaults.fill_params_host(self.params_host)
        
    def fill(self):
        self.fill_settings()
        self.fill_params()
        self.fill_params_host()
        
        
    #==========================================================================
    #::: example initializations
    #==========================================================================
    def init_hot_jupiter(self):
        self.settings = defaults.get_hot_jupiter_settings()
        self.params = defaults.get_hot_jupiter_params()
        self.params_host = defaults.get_hot_jupiter_params_host()
        self.fill()
        
        
    #==========================================================================
    #::: add stuff
    #==========================================================================
    def add_companion_phot(self, name):
        if 'companions_phot' not in self.settings: self.settings['companions_phot'] = []
        self.settings['companions_phot'] += [name]
        self.fill()
        
    def add_companion_rv(self, name):
        if 'companions_rv' not in self.settings: self.settings['companions_rv'] = []
        self.settings['companions_rv'] += [name]
        self.fill()
        
    def add_inst_phot(self, name):
        if 'inst_phot' not in self.settings: self.settings['inst_phot'] = []
        self.settings['inst_phot'] += [name]
        self.fill()
        
    def add_inst_rv(self, name):
        if 'inst_rv' not in self.settings: self.settings['inst_rv'] = []
        self.settings['inst_rv'] += [name]
        self.fill()
        
    def add_flare(self, name):
        pass #TODO
    
    def add_observation(self, inst, key, time, y, yerr):
        self.data[inst] = {}
        self.data[inst]['time'] = time
        self.data[inst][key] = y
        self.data[inst][key+'_err'] = yerr
        
        
    #==========================================================================
    #::: do stuff
    #==========================================================================
    def generate_model(self, xx, inst='telescope', key='flux'):
        params_ellc = translator.translate_alles_to_ellc(self.params, self.settings)
        model = computer.calculate_model(params_ellc, inst, key, xx=xx, settings=self.settings)
        return model
        
    # def plot_model(self, inst, key, xx):
    #     model = self.generate_model(time, inst, key)
    #     pass
        
    # def inject_flux_model(self, inst, **kwargs):
    #     # generator.inject_transit_model(self.data[inst]['time'], self.data[inst]['flux'], self.data[inst]['flux_err'], params=self.params, settings=self.settings, **kwargs)
    #     pass
        
        
# class host():
    
#     def __init__(self, params=None, settings=None):
#         self.params = {'R_host': R_host,
#                        'M_host': M_host}

    

# class companion():
    
#     def __init__(self, name):
#         self.name = name
#         self.params = {'R_companion_earth': R_companion_earth,
#                        'M_companion_earth': M_companion_earth}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 14:28:55 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter, MaxNLocator
import os
import gzip
try:
   import cPickle as pickle
except:
   import pickle
from copy import deepcopy
from dynesty import utils as dyutils
from dynesty import plotting as dyplot
import warnings

#::: allesfitter modules
from . import config
from . import deriver
from .computer import calculate_model, calculate_baseline, calculate_stellar_var
from .general_output import afplot, afplot_per_transit, save_table, save_latex_table, logprint, get_params_from_samples, plot_ttv_results
from .plot_top_down_view import plot_top_down_view
from .utils.colormaputil import truncate_colormap
from .utils.latex_printer import round_tex
from .statistics import residual_stats
                     

    

###############################################################################
#::: draw samples from the ns results (internally in the code)
###############################################################################
def draw_ns_posterior_samples(results, Nsamples=None, as_type='2d_array'):
    '''
    ! posterior samples are drawn as resampled weighted samples !
    ! do not confuse posterior_samples (weighted, resampled) with results['samples'] (unweighted) !
    '''
    weights = np.exp(results['logwt'] - results['logz'][-1])
    np.random.seed(42)
    posterior_samples = dyutils.resample_equal(results['samples'], weights)    
    if Nsamples:
        posterior_samples = posterior_samples[np.random.randint(len(posterior_samples), size=Nsamples)]

    if as_type=='2d_array':
        return posterior_samples
    
    elif as_type=='dic':
        posterior_samples_dic = {}
        for key in config.BASEMENT.fitkeys:
            ind = np.where(config.BASEMENT.fitkeys==key)[0]
            posterior_samples_dic[key] = posterior_samples[:,ind].flatten()
        return posterior_samples_dic



###############################################################################
#::: analyse the output from save_ns.pickle file
###############################################################################
def ns_output(datadir):
    '''
    Inputs:
    -------
    datadir : str
        the working directory for allesfitter
        must contain all the data files
        output directories and files will also be created inside datadir
            
    Outputs:
    --------
    This will output information into the console, and create a output files 
    into datadir/results/ (or datadir/QL/ if QL==True)    
    '''
    config.init(datadir)
    
    
    #::: security check
    if os.path.exists(os.path.join(config.BASEMENT.outdir,'ns_table.csv')):
        try:
            overwrite = str(input('Nested Sampling output files already exists in '+config.BASEMENT.outdir+'\n'+\
                                  '-----------------------------------------------'+''.join(['-']*len(config.BASEMENT.outdir))+'\n'\
                                  'What do you want to do?\n'+\
                                  '1 : overwrite the output files\n'+\
                                  '2 : abort\n'))
            if (overwrite == '1'):
                print('\n')
                pass
            else:
                raise ValueError('User aborted operation.')
        except EOFError:
            warnings.warn("Nested Sampling output files already existed from a previous run, and were automatically overwritten.")
            pass
    
    
    #::: load the save_ns.pickle
    f = gzip.GzipFile(os.path.join(config.BASEMENT.outdir,'save_ns.pickle.gz'), 'rb')
    results = pickle.load(f)
    f.close()
           
    
    #::: plot the fit
    posterior_samples_for_plot = draw_ns_posterior_samples(results, Nsamples=20) #only 20 samples for plotting
    
    for companion in config.BASEMENT.settings['companions_all']:
        fig, axes = afplot(posterior_samples_for_plot, companion)
        if fig is not None:
            fig.savefig( os.path.join(config.BASEMENT.outdir,'ns_fit_'+companion+'.pdf'), bbox_inches='tight' )       
            plt.close(fig)

    for companion in config.BASEMENT.settings['companions_phot']:
        for inst in config.BASEMENT.settings['inst_phot']:
            fig, axes = afplot_per_transit(posterior_samples_for_plot, inst, companion)
            fig.savefig( os.path.join(config.BASEMENT.outdir,'ns_fit_per_transit_'+inst+'_'+companion+'.pdf'), bbox_inches='tight' )
            plt.close(fig)
    
    
    #::: retrieve the results
    posterior_samples = draw_ns_posterior_samples(results)                               # all weighted posterior_samples
    params_median, params_ll, params_ul = get_params_from_samples(posterior_samples)     # params drawn form these posterior_samples
    
    
    #::: output the results
    logprint('\nResults:')
    logprint('----------')
#    print(results.summary())
    logZdynesty = results.logz[-1]                                                       # value of logZ
    logZerrdynesty = results.logzerr[-1]                                                 # estimate of the statistcal uncertainty on logZ
    logprint('log(Z) = {} +- {}'.format(logZdynesty, logZerrdynesty))
    logprint('Nr. of posterior samples: {}'.format(len(posterior_samples)))
    
    
    #::: make pretty titles for the plots  
    labels, units = [], []
    for i,l in enumerate(config.BASEMENT.fitlabels):
        labels.append( str(config.BASEMENT.fitlabels[i]) )
        units.append( str(config.BASEMENT.fitunits[i]) )
        
    results2 = deepcopy(results) # results.copy() does not work anymore since dynesty 1.2                 
    params_median2, params_ll2, params_ul2 = params_median.copy(), params_ll.copy(), params_ul.copy()     # params drawn form these posterior_samples; only needed for plots (subtract epoch offset)  
    fittruths2 = config.BASEMENT.fittruths.copy()
    for companion in config.BASEMENT.settings['companions_all']:
        
        if companion+'_epoch' in config.BASEMENT.fitkeys:
            ind = np.where(config.BASEMENT.fitkeys==companion+'_epoch')[0][0]
            results2['samples'][:,ind] -= int(params_median[companion+'_epoch'])                #np.round(params_median[companion+'_epoch'],decimals=0)
            units[ind] = str(units[ind]+'-'+str(int(params_median[companion+'_epoch']))+'d')    #np.format_float_positional(params_median[companion+'_epoch'],0)+'d')
            fittruths2[ind] -= int(params_median[companion+'_epoch'])
            params_median2[companion+'_epoch'] -= int(params_median[companion+'_epoch'])
                

    for i,l in enumerate(labels):
        if len( units[i].strip(' ') ) > 0:
            labels[i] = str(labels[i]+' ('+units[i]+')')
        
        
    #::: traceplot    
    cmap = truncate_colormap( 'Greys', minval=0.2, maxval=0.8, n=256 )
    tfig, taxes = dyplot.traceplot(results2, labels=labels, quantiles=[0.16, 0.5, 0.84], truths=fittruths2, post_color='grey', trace_cmap=[cmap]*config.BASEMENT.ndim, trace_kwargs={'rasterized':True})
    plt.tight_layout()
    
    
    #::: cornerplot
    # ndim = results2['samples'].shape[1]
    fontsize = np.min(( 24. + 0.5*config.BASEMENT.ndim, 40 ))
    try:
        cfig, caxes = dyplot.cornerplot(results2, labels=labels, span=[0.997 for i in range(config.BASEMENT.ndim)], quantiles=[0.16, 0.5, 0.84], truths=fittruths2, hist_kwargs={'alpha':0.25,'linewidth':0,'histtype':'stepfilled'}, 
                                        label_kwargs={"fontsize":fontsize, "rotation":45, "horizontalalignment":'right'})
    except:
        logprint('! WARNING')
        logprint('Dynesty corner plot could not be created. Please contact maxgue@mit.edu.')
        cfig, caxes = plt.subplots(config.BASEMENT.ndim,config.BASEMENT.ndim,figsize=(2*config.BASEMENT.ndim,2*config.BASEMENT.ndim))
        
        
    #::: runplot
#    rfig, raxes = dyplot.runplot(results)
#    rfig.savefig( os.path.join(config.BASEMENT.outdir,'ns_run.jpg'), dpi=100, bbox_inches='tight' )
#    plt.close(rfig)
    
    
    #::: set allesfitter titles and labels
    for i, key in enumerate(config.BASEMENT.fitkeys): 
        
        value = round_tex(params_median2[key], params_ll2[key], params_ul2[key])
        ctitle = r'' + labels[i] + '\n' + r'$=' + value + '$'
        ttitle = r'' + labels[i] + r'$=' + value + '$'
        if len(config.BASEMENT.fitkeys)>1:
            # caxes[i,i].set_title(ctitle)
            caxes[i,i].set_title(ctitle, fontsize=fontsize, rotation=45, horizontalalignment='left')
            taxes[i,1].set_title(ttitle)
            for i in range(caxes.shape[0]):
                for j in range(caxes.shape[1]):
                    caxes[i,j].xaxis.set_label_coords(0.5, -0.5)
                    caxes[i,j].yaxis.set_label_coords(-0.5, 0.5)
        
                    if i==(caxes.shape[0]-1): 
                        fmt = ScalarFormatter(useOffset=False)
                        caxes[i,j].xaxis.set_major_locator(MaxNLocator(nbins=3))
                        caxes[i,j].xaxis.set_major_formatter(fmt)
                    if (i>0) and (j==0):
                        fmt = ScalarFormatter(useOffset=False)
                        caxes[i,j].yaxis.set_major_locator(MaxNLocator(nbins=3))
                        caxes[i,j].yaxis.set_major_formatter(fmt)
                        
                    for tick in caxes[i,j].xaxis.get_major_ticks(): tick.label.set_fontsize(24) 
                    for tick in caxes[i,j].yaxis.get_major_ticks(): tick.label.set_fontsize(24)    
        else:
            caxes[i,i].set_title(ctitle)
            taxes[1].set_title(ttitle)
            caxes[i,i].xaxis.set_label_coords(0.5, -0.5)
            caxes[i,i].yaxis.set_label_coords(-0.5, 0.5)
               
            
    #::: save and close the trace- and cornerplot
    tfig.savefig( os.path.join(config.BASEMENT.outdir,'ns_trace.pdf'), bbox_inches='tight' )
    plt.close(tfig)
    cfig.savefig( os.path.join(config.BASEMENT.outdir,'ns_corner.pdf'), bbox_inches='tight' )
    plt.close(cfig)


    #::: save the tables
    save_table(posterior_samples, 'ns')
    save_latex_table(posterior_samples, 'ns')
    

    #::: derive values (using stellar parameters from params_star.csv)
    deriver.derive(posterior_samples, 'ns')
    
    
    #::: check the residuals
    for inst in config.BASEMENT.settings['inst_all']:
        if inst in config.BASEMENT.settings['inst_phot']: key='flux'
        elif inst in config.BASEMENT.settings['inst_rv']: key='rv'
        model = calculate_model(params_median, inst, key)
        baseline = calculate_baseline(params_median, inst, key)
        stellar_var = calculate_stellar_var(params_median, inst, key)
        residuals = config.BASEMENT.data[inst][key] - model - baseline - stellar_var
        residual_stats(residuals)
    
    
    #::: make top-down orbit plot (using stellar parameters from params_star.csv)
    try:
        params_star = np.genfromtxt( os.path.join(config.BASEMENT.datadir,'params_star.csv'), delimiter=',', names=True, dtype=None, encoding='utf-8', comments='#' )
        fig, ax = plot_top_down_view(params_median, params_star)
        fig.savefig( os.path.join(config.BASEMENT.outdir,'top_down_view.pdf'), bbox_inches='tight' )
        plt.close(fig)        
    except:
        logprint('\nOrbital plots could not be produced.')
    
    
    #::: plot TTV results (if wished for)
    if config.BASEMENT.settings['fit_ttvs'] == True:
        plot_ttv_results(params_median, params_ll, params_ul)
    
    
    #::: clean up
    logprint('\nDone. For all outputs, see', config.BASEMENT.outdir)
    
    
    #::: return a nerdy quote
    try:
        with open(os.path.join(os.path.dirname(__file__), 'utils', 'quotes.txt')) as dataset:
            return(np.random.choice([l for l in dataset]))
    except:
        return('42')
    
    

def ns_derive(datadir): #emergency function if matplotlib and Mac OSX crash
    posterior_samples = get_ns_posterior_samples(datadir, as_type='2d_array')
    deriver.derive(posterior_samples, 'ns')
    
    
    
###############################################################################
#::: get NS samples (for top-level user)
###############################################################################
def get_ns_posterior_samples(datadir, Nsamples=None, as_type='dic'):
    config.init(datadir)
    
    try:
        f = gzip.GzipFile(os.path.join(datadir,'results','save_ns.pickle.gz'), 'rb')
        results = pickle.load(f)
        f.close()
        
    except:
        with open(os.path.join(datadir,'results','save_ns.pickle'),'rb') as f:
            results = pickle.load(f)    

    return draw_ns_posterior_samples(results, Nsamples=Nsamples, as_type=as_type)
    
    
 
###############################################################################
#::: get NS params (for top-level user)
###############################################################################   
def get_ns_params(datadir):
    posterior_samples = get_ns_posterior_samples(datadir, Nsamples=None, as_type='2d_array')  # all weighted posterior_samples
    params_median, params_ll, params_ul = get_params_from_samples(posterior_samples)     # params drawn form these posterior_samples
    return params_median, params_ll, params_ul


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 13:37:00 2020

@author: 
Dr. Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
GitHub: https://github.com/MNGuenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import os
import ellc
from astropy.constants import G
from astropy import units as u

#::: my modules
from allesfitter.exoworlds_rdx.lightcurves import lightcurve_tools as lct

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})





###############################################################################
#::: translate physical parameters to ellc parameters
###############################################################################
def translate_physical_to_ellc(period = 1., 
                               R_companion = 1., M_companion = 1.,
                               R_companion_unit = 'Rearth', M_companion_unit = 'Mearth',
                               R_host = 1., M_host = 1., 
                               incl = 90, 
                               ecc = 0,
                               omega = 0):
    
    if R_companion_unit=='Rearth':
        R_companion_with_unit = R_companion*u.Rearth
    elif R_companion_unit=='Rjup':
        R_companion_with_unit = R_companion*u.Rjup
    elif R_companion_unit=='Rsun':
        R_companion_with_unit = R_companion*u.Rsun
        
    if M_companion_unit=='Mearth':
        M_companion_with_unit = M_companion*u.Mearth
    elif M_companion_unit=='Mjup':
        M_companion_with_unit = M_companion*u.Mjup
    elif R_companion_unit=='Msun':
        R_companion_with_unit = R_companion*u.Msun
    
    a = (G/(4*np.pi**2) * (period*u.d)**2 * (M_host*u.Msun + M_companion_with_unit))**(1./3.) #in AU, with astropy units
    a = a.to(u.AU) #in AU, with astropy units
    
    cosi = np.cos(np.deg2rad(incl))
    
    f_c = np.sqrt(ecc) * np.cos(np.deg2rad(omega))
    f_s = np.sqrt(ecc) * np.sin(np.deg2rad(omega))
    
    radius_1 = (R_host*u.Rsun / a).decompose().value
    radius_2 = (R_companion_with_unit / a).decompose().value
    a = a.value    

    return radius_1, radius_2, cosi, a, f_c, f_s



###############################################################################
#::: translate physical parameters to ellc parameters
###############################################################################
def translate_physical_to_allesfitter(period = 1., 
                                      R_companion = 1., M_companion = 1.,
                                      R_companion_unit = 'Rearth', M_companion_unit = 'Mearth',
                                      R_host = 1., M_host = 1., 
                                      incl = 90, 
                                      ecc = 0,
                                      omega = 0,
                                      ldc = [0.6,0.2],
                                      ld = 'quad'):
    
    if R_companion_unit=='Rearth':
        R_companion_with_unit = R_companion*u.Rearth
    elif R_companion_unit=='Rjup':
        R_companion_with_unit = R_companion*u.Rjup
    elif R_companion_unit=='Rsun':
        R_companion_with_unit = R_companion*u.Rsun
        
    if M_companion_unit=='Mearth':
        M_companion_with_unit = M_companion*u.Mearth
    elif M_companion_unit=='Mjup':
        M_companion_with_unit = M_companion*u.Mjup
    elif R_companion_unit=='Msun':
        R_companion_with_unit = R_companion*u.Msun
    
    cosi = np.cos(np.deg2rad(incl))
    
    a = (G/(4*np.pi**2) * (period*u.d)**2 * (M_host*u.Msun + M_companion_with_unit))**(1./3.) #in AU, with astropy units
    a = a.to(u.AU) #in AU, with astropy units
    
    f_c = np.sqrt(ecc) * np.cos(np.deg2rad(omega))
    f_s = np.sqrt(ecc) * np.sin(np.deg2rad(omega))
    
    if ld == 'quad':
        q1 = (ldc[0] + ldc[1])**2
        q2 = 0.5 * ldc[0] / (ldc[0] + ldc[1])
        ldc_transformed = [q1, q2]
        
    rr = ( R_companion_with_unit / (R_host*u.Rsun) ).decompose().value
    rsuma = ( (R_host*u.Rsun + R_companion_with_unit) / a ).decompose().value
    a = a.value    

    return rr, rsuma, cosi, a, f_c, f_s, ldc_transformed



###############################################################################
#::: make an ellc model lightcurve
###############################################################################
def make_lc_model(time, 
                  flux = None, flux_err = None,
                  epoch = 0., period = 1., 
                  R_companion = 1., M_companion = 1.,
                  R_companion_unit = 'Rearth', M_companion_unit = 'Mearth',
                  R_host = 1., M_host = 1., 
                  incl = 90, 
                  ecc = 0,
                  omega = 0,
                  ldc = [0.6,0.2],
                  ld = 'quad',
                  dil = 0,
                  sbratio = 0,
                  show_plot = False, save_plot = False, save_csv = False,
                  fname_plot = 'lc.pdf', fname_csv = 'lc.csv'):
    '''
    Inputs:
    -------
    time : array of float
        time in days
        
    Optional Inputs:
    ----------------
    flux : array of float
        flux of the 'underlying' lightcurve
    flux_err : array of float
        flux error of the 'underlying' lightcurve
    epoch : float
        epoch in days
    period : float
        period in days
    R_companion: float
        radius of the companion
        default is 1 Rearth
    M_companion: float
        mass of the companion
        default is 1 Mearth
    R_companion: float
        radius of the companion
        default is 1 Rearth
    M_companion: float
        mass of the companion
        default is 1 Mearth
    R_host : float
        radius of the star, in Rsun
        default is 1
    M_host: float
        mass of the star, in Msun
        default is 1
    show_plot : bool
        show the plot in the terminal, or close it
        default is False
    save_plot : bool
        save the plot to a file, or not
        default is False
    save_csv : bool
        save the lightcurve to a file, or not
        default is False
        
    Returns:
    --------
    model_flux : array of float
        relative flux of the model
    '''
    
    radius_1, radius_2, cosi, a, f_c, f_s = translate_physical_to_ellc(period = period, R_companion = R_companion, M_companion = M_companion, R_companion_unit = R_companion_unit, M_companion_unit = M_companion_unit, R_host = R_host, M_host = M_host, incl = incl, ecc = ecc, omega = omega)   
    rr, rsuma, cosi, a, f_c, f_s, ldc_transformed = translate_physical_to_allesfitter(period = period, R_companion = R_companion, M_companion = M_companion, R_companion_unit = R_companion_unit, M_companion_unit = M_companion_unit, R_host = R_host, M_host = M_host, incl = incl, ecc = ecc, omega = omega)   

    model_flux = ellc.lc(t_obs = time, radius_1 = radius_1, radius_2 = radius_2, sbratio = sbratio, incl = incl, light_3 = dil, t_zero = epoch, period = period, a = a, q = 1, f_c = f_c, f_s = f_s, ldc_1=ldc, ldc_2 = None, gdc_1 = None, gdc_2 = None, didt = None, domdt = None, rotfac_1 = 1, rotfac_2 = 1, hf_1 = 1.5, hf_2 = 1.5, bfac_1 = None, bfac_2 = None, heat_1 = None, heat_2 = None, lambda_1 = None, lambda_2 = None, vsini_1 = None, vsini_2 = None, t_exp=None, n_int=None,  grid_1='default', grid_2='default', ld_1=ld, ld_2=None, shape_1='sphere', shape_2='sphere', spots_1=None, spots_2=None, exact_grav=False, verbose=1)
    
    if flux is not None:
        flux2 = flux+model_flux-1
    else:
        flux2 = model_flux
        
    if show_plot or save_plot:        
        
        #::: get phase-folded data
        phase, phaseflux, phaseflux_err, N, phi = lct.phase_fold(time, flux2, period, epoch, dt=0.002, ferr_type='medsig', ferr_style='sem', sigmaclip=True)

        #::: get model on fine grid from phase -0.25 to 0.75
        xx = np.linspace( epoch-0.25*period, epoch+0.75*period, 1000 )
        xx_phase = np.linspace(-0.25, 0.75, 1000 )
        model_flux_xx_phase = ellc.lc(t_obs = xx, radius_1 = radius_1, radius_2 = radius_2, sbratio = sbratio, incl = incl, light_3 = dil, t_zero = epoch, period = period, a = a, q = 1, f_c = f_c, f_s = f_s, ldc_1=ldc, ldc_2 = None, gdc_1 = None, gdc_2 = None, didt = None, domdt = None, rotfac_1 = 1, rotfac_2 = 1, hf_1 = 1.5, hf_2 = 1.5, bfac_1 = None, bfac_2 = None, heat_1 = None, heat_2 = None, lambda_1 = None, lambda_2 = None, vsini_1 = None, vsini_2 = None, t_exp=None, n_int=None,  grid_1='default', grid_2='default', ld_1=ld, ld_2=None, shape_1='sphere', shape_2='sphere', spots_1=None, spots_2=None, exact_grav=False, verbose=1)
            
        fig = plt.figure(figsize=(15,10), tight_layout=True)
        gs = gridspec.GridSpec(3, 3)
        
        #::: full curve
        ax = fig.add_subplot(gs[0,:])
        ax.plot(time, flux2, 'b.', rasterized=True, zorder=10)
        ax.plot(time, model_flux, 'r-', lw=2, zorder=12)
        ax.set(ylabel='Flux', xlabel='Time'+r' $\mathrm{(BJD_{TDB})}$')
        
        #::: phase-folded curve
        ax = fig.add_subplot(gs[1,0])
        ax.plot(phi, flux2, 'k.', color='lightgrey', rasterized=True, zorder=10)
        ax.errorbar(phase, phaseflux, yerr=phaseflux_err, fmt='b.', zorder=11)
        ax.plot(xx_phase, model_flux_xx_phase, 'r-', lw=2, zorder=12)
        ax.set(xlabel='Phase', ylabel='Flux')
        
        #::: phase-folded and zoomed curve; same as for axes[0,1], but zoomed in x
        ax = fig.add_subplot(gs[1,1])
        ax.plot(phi*period*24.*60, flux2, 'k.', color='lightgrey', rasterized=True, zorder=10)
        ax.errorbar(phase*period*24.*60, phaseflux, yerr=phaseflux_err, fmt='b.', zorder=11)
        ax.plot(xx_phase*period*24.*60, model_flux_xx_phase, 'r-', lw=2, zorder=12)
        ax.set(xlim=[-0.03*period*24.*60,0.03*period*24.*60])
        ax.set(xlabel='Time (min.)')
        
        #::: info
        ax = fig.add_subplot(gs[2,0])
        ax.set_axis_off()
        ax.text(0,0.95,r'$T_0$' + ' = '+str(epoch)+r' $\mathrm{BJD_{TDB}}$', transform=ax.transAxes)
        ax.text(0,0.85,'P = '+str(period)+' days', transform=ax.transAxes)
        ax.text(0,0.75,'R_comp = '+str(R_companion)+' '+R_companion_unit, transform=ax.transAxes)
        ax.text(0,0.65,'M_comp = '+str(M_companion)+' '+M_companion_unit, transform=ax.transAxes)
        ax.text(0,0.55,'R_host = '+str(R_host)+' Rsun', transform=ax.transAxes)
        ax.text(0,0.45,'M_host = '+str(M_host)+' Msun', transform=ax.transAxes)
        ax.text(0,0.35,'incl = '+str(incl)+' deg', transform=ax.transAxes)
        ax.text(0,0.25,'ecc = '+str(ecc), transform=ax.transAxes)
        ax.text(0,0.15,'omega = '+str(omega)+' deg', transform=ax.transAxes)
        ax.text(0,0.05,'LD = '+str(ldc)+' '+ld, transform=ax.transAxes)
        
        ax = fig.add_subplot(gs[2,1])
        ax.set_axis_off()
        ax.text(0,0.95,'dil = '+str(dil), transform=ax.transAxes)
        ax.text(0,0.85,'sbratio = '+str(sbratio), transform=ax.transAxes)
        ax.text(0,0.75,'R_comp/R_host = '+np.format_float_positional(rr,5,False), transform=ax.transAxes)
        ax.text(0,0.65,'(R_comp+R_host)/a = '+np.format_float_positional(rsuma,5,False), transform=ax.transAxes)
        ax.text(0,0.55,'R_comp/a = '+np.format_float_positional(radius_2,5,False), transform=ax.transAxes)
        ax.text(0,0.45,'R_host/a = '+np.format_float_positional(radius_1,5,False), transform=ax.transAxes)
        ax.text(0,0.35,'cosi = '+np.format_float_positional(cosi,5,False), transform=ax.transAxes)
        ax.text(0,0.25,r'$\sqrt{e} \cos{\omega}$ = '+np.format_float_positional(f_c,5,False), transform=ax.transAxes)
        ax.text(0,0.15,r'$\sqrt{e} \sin{\omega}$ = '+np.format_float_positional(f_s,5,False), transform=ax.transAxes)
        ax.text(0,0.05,'LD transf = ['+", ".join([np.format_float_positional(item,5,False) for item in ldc_transformed]) + ']', transform=ax.transAxes)
        
        if show_plot:
            plt.show(fig)
            
        if save_plot:
            if len(os.path.dirname(fname_plot))>0 and not os.path.exists(os.path.dirname(fname_plot)): os.makedirs(os.path.dirname(fname_plot))
            fig.savefig(fname_plot)
            plt.close(fig)
    
    if save_csv:
        if len(os.path.dirname(fname_csv))>0 and not os.path.exists(os.path.dirname(fname_csv)): os.makedirs(os.path.dirname(fname_csv))
        X = np.column_stack((time, flux2, flux_err))
        np.savetxt(fname_csv, X, delimiter=',')


    return flux2

    
    
###############################################################################
#::: inject an ellc model lightcurve
###############################################################################
def inject_lc_model(time, flux, flux_err, **kwargs):
    '''
    Wrapper around make_lc_model()
    '''
    
    return make_lc_model(time, flux = flux, flux_err = flux_err, **kwargs)
    


   
###############################################################################
#::: make an ellc model lightcurve
###############################################################################
def inject_rv_model(time, rv, rv_err, 
               epoch, period, 
               R_companion= 1., M_companion=1., 
               R_companion_unit = 'Rearth', M_companion_unit = 'Mearth',
               R_host = 1., M_host = 1., 
               sbratio = 0,
               incl = 90, 
               ecc = 0,
               omega = 0,
               dil = 0,
               ldc = [0.6,0.2],
               ld = 'quad',
               show_plot = False, save_plot = False, save_csv = False,
               outdir = '', fname_plot=None, fname_csv=None):
    '''
    Inputs:
    -------
    time : array of float
        time in days
    rv : array of float
        RV of the 'underlying' series
    rv_err : array of float
        error of RV of the 'underlying' series
    epoch : float
        epoch in days
    period : float
        period in days
    R_companion: float
        radius of the planet, in R_earth
        
    Optional Inputs:
    ----------------
    M_companion: float
        mass of the planet, in M_earth
        default is 0
    R_host : float
        radius of the star, in R_sun
        default is 1
    M_host: float
        mass of the star, in M_sun
        default is 1
    show_plot : bool
        show the plot in the terminal, or close it
        default is False
    save_plot : bool
        save the plot to a file, or not
        default is False
    save_csv : bool
        save the lightcurve to a file, or not
        default is False
        
    Returns:
    --------
    flux2 : array of float
        relative flux with injected signal
    '''
    
    if R_companion_unit=='Rearth':
        R_companion_with_unit = R_companion*u.Rearth
    elif R_companion_unit=='Rjup':
        R_companion_with_unit = R_companion*u.Rjup
        
    if M_companion_unit=='Mearth':
        M_companion_with_unit = M_companion*u.Mearth
    elif M_companion_unit=='Mjup':
        M_companion_with_unit = M_companion*u.Mjup
    
    a = (G/(4*np.pi**2) * (period*u.d)**2 * (M_host*u.Msun + M_companion_with_unit))**(1./3.) #in AU, with astropy units
    a = a.to(u.AU) #in AU, with astropy units
    
    if ecc>0:
        f_c = np.sqrt(ecc) * np.cos(np.deg2rad(omega))
        f_s = np.sqrt(ecc) * np.sin(np.deg2rad(omega))
    else:
        f_c = None
        f_s = None
        
    radius_1 = (R_host*u.Rsun / a).decompose().value
    radius_2 = (R_companion_with_unit / a).decompose().value
    a = a.value
    
    def get_model_rv(xx):
        return ellc.rv(t_obs = xx, 
                       radius_1 = radius_1,
                       radius_2 = radius_2,
                       sbratio = sbratio, 
                       incl = incl, 
                       t_zero = epoch, 
                       period = period,
                       a = a,
                       q = 1,
                       f_c = f_c, f_s = f_s,
                       ldc_1=ldc, ldc_2 = None,
                       gdc_1 = None, gdc_2 = None,
                       didt = None,
                       domdt = None,
                       rotfac_1 = 1, rotfac_2 = 1,
                       hf_1 = 1.5, hf_2 = 1.5,
                       bfac_1 = None, bfac_2 = None,
                       heat_1 = None, heat_2 = None, 
                       lambda_1 = None, lambda_2 = None,
                       vsini_1 = None, vsini_2 = None,
                       t_exp=None, n_int=None, 
                       grid_1='default', grid_2='default',
                       ld_1=ld, ld_2=None,
                       shape_1='sphere', shape_2='sphere',
                       spots_1=None, spots_2=None, 
                       verbose=1)[0]
    
    #::: get model on data grid
    model_rv = get_model_rv(time)
    rv2 = rv+model_rv
    
    if show_plot or save_plot:
        
        #::: get phase 
        phi = lct.calc_phase(time, period, epoch)
        phi[phi>0.75] -= 1.

        #::: get model on fine grid
        xx1 = np.arange( time[0], time[-1], 0.1 )
        model_rv_xx1 = get_model_rv(xx1)
    
        #::: get model on fine grid from phase -0.25 to 0.75
        xx = np.linspace( epoch-0.25*period, epoch+0.75*period, 1000 )
        xx_phase = np.linspace(-0.25, 0.75, 1000 )
        model_rv_xx_phase = get_model_rv(xx)
    
        fig, axes = plt.subplots(1, 2, figsize=(16,5), sharey='row', gridspec_kw={'width_ratios': [3,1]})
        
        #::: full curve
        axes[0].errorbar(time, rv2, yerr=rv_err, fmt='bo', zorder=11)
        axes[0].plot(xx1, model_rv_xx1, 'r-', lw=2, zorder=10)
        axes[0].set(ylabel='RV (km/s)', xlabel='Time (BJD_TDB)')
        
        #::: phase-folded curve
        axes[1].errorbar(phi, rv2, yerr=rv_err, fmt='b.', zorder=11)
        axes[1].plot(xx_phase, model_rv_xx_phase, 'r-', lw=2, zorder=10)
        axes[1].set(xlabel='Phase')
        
        suptitle = fig.suptitle('\nP = '+str(period)+' days, Rp = '+str(R_companion)+' '+R_companion_unit, y=1.15)
        fig.tight_layout()        

        if show_plot:
            plt.show(fig)
            
        if save_plot:
            if fname_plot is None: 
                if not os.path.exists(os.path.join(outdir,'plots')): os.makedirs(os.path.join(outdir,'plots'))
                fname_plot = os.path.join(outdir,'plots','Lightcurve P = '+str(period)+' days, Rp = '+str(R_companion)+' '+R_companion_unit+'.pdf')
            fig.savefig(fname_plot, bbox_extra_artists=(suptitle,), bbox_inches="tight")
            plt.close(fig)
    
    if save_csv:
        if fname_csv is None: 
            if not os.path.exists(os.path.join(outdir,'csv')): os.makedirs(os.path.join(outdir,'csv'))
            fname_csv = os.path.join(outdir,'csv','Lightcurve P = '+str(period)+' days, Rp = '+str(R_companion)+' '+R_companion_unit+'.csv')
        X = np.column_stack((time, rv2, rv_err))
        np.savetxt(fname_csv, X, delimiter=',')


    return rv2


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 11 17:32:25 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os, sys
from datetime import datetime
import emcee
import corner
import multiprocessing
multiprocessing.set_start_method('fork', force=True)
#solves python>=3.8 issues, see https://stackoverflow.com/questions/60518386/error-with-module-multiprocessing-under-python3-8
from multiprocessing import Pool, cpu_count
from contextlib import closing

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

np.random.seed(21)




###############################################################################
#::: set up MCMC log probability function
#::: (has to be top-level for pickle)
###############################################################################
def log_probability(params):
    '''
    works on X, Y
    '''
    
    try:
        ll = log_likelihood(params)
        lp = external_log_prior(params)
    except:
        return -np.inf
    if not np.isfinite(lp):
        return -np.inf
    return ll + lp



###############################################################################
#::: priors
###############################################################################  
def external_log_prior(params):
    log_jitter = params
    
    lp = 0
    if not (-23 < log_jitter < 0):
        lp = -np.inf
    
    return lp

    

###############################################################################
#::: likelihood
###############################################################################          
def log_likelihood(theta):
    log_jitter = theta
    jitter = np.exp(log_jitter)
    yerr = np.sqrt( WHITE_NOISE**2 + jitter**2 )
    inv_sigma2_w = 1./yerr**2
    return -0.5*(np.nansum((Y)**2 * inv_sigma2_w - np.log(inv_sigma2_w)))
        


###############################################################################
#::: run
###############################################################################
def estimate_jitter(x,y,white_noise,
        jitter_guess=None,
        mean=0.,
        nwalkers=50, thin_by=50, burn_steps=2500, total_steps=5000,
        bin_width=None,
        xlabel='x', ylabel='y', ydetr_label='ydetr',
        outdir='jitter_fit', fname=None, fname_summary=None,
        multiprocess=False, multiprocess_cores=None):
    
    '''
    Required Input:
    ---------------
    x : array of float
        x-values of the data set
    y : array of float
        y-values of the data set
    white_noise : array of float / float
        white_noise on y-values of the data set
        
    Optional Input:
    ---------------
    mean : float (default 0.)
        mean of the data set
        the default is 1., assuming usually y will be normalized flux
    nwalkers : int
        number of MCMC walkers
    thin_by : int
        thinning the MCMC chain by how much
    burn_steps : int
        how many steps to burn in the MCMC
    total_steps : int
        total MCMC steps (including burn_steps)
    xlabel : str
        x axis label (for plots)
    ylabel : str
        y axis label (for plots)       
    ydetr_label : str
        y_detr axis label (for plots)    
    outdir : str
        name of the output directory
    fname : str
        prefix of the output files (e.g. a planet name)
    multiprocess : bool (default False)
        run MCMC on multiprocess_cores cores        
    multiprocess_cores : bool (default None)
        run MCMC on many cores        
    '''
    
    #::: this is ugly, I know;
    #::: blame the multiprocessing and pickling issues, 
    #::: which demand global variables for efficiency    
    global X
    global Y
    global WHITE_NOISE
    global MEAN
    X = x
    Y = y
    WHITE_NOISE = white_noise
    MEAN = mean
    

    #::: outdir
    if not os.path.exists(outdir): os.makedirs(outdir)
    
    
    #::: print function that prints into console and logfile at the same time 
    now = datetime.now().isoformat()
    def logprint(*text):
        print(*text)
        original = sys.stdout
        with open( os.path.join(outdir,fname+'logfile_'+now+'.log'), 'a' ) as f:
            sys.stdout = f
            print(*text)
        sys.stdout = original

    
    #::: fname
    if fname is not None:
        fname += '_jitter_fit_'
    else:
        fname = 'jitter_fit_'

    
    #::: MCMC plot settings
    names = [r'$\log{(y_\mathrm{err})}$']
    discard = int(1.*burn_steps/thin_by)


    logprint('\nStarting...')


    #guess the total yerr = np.sqrt( rv_err**2 + jitter**2  )
    yerr = np.nanstd(Y) * np.ones_like(Y)


    #re-calcualte an initital guess for log_jitter 
    if jitter_guess is None:
        jitter_guess = np.nanmedian( np.log( np.sqrt( yerr**2 - white_noise**2 ) ) )


    #::: plot the data
    fig, ax = plt.subplots()
    ax.errorbar(x, y, yerr=white_noise, fmt=".b", capsize=0)
    ax.set( xlabel=xlabel, ylabel=ylabel, title='Original data' )
    fig.savefig( os.path.join(outdir,fname+'data.jpg'), dpi=100, bbox_inches='tight')
    plt.close(fig)
    

    
    
    ###########################################################################
    #::: MCMC fit
    ###########################################################################
    if multiprocess and not multiprocess_cores:
        multiprocess_cores = cpu_count()-1
        
    logprint('\nRunning MCMC fit...')
    if multiprocess: logprint('\tRunning on', multiprocess_cores, 'CPUs.')   
    
    
    #::: all
    initial = np.array([jitter_guess])

    
    #::: set up MCMC
    ndim = len(initial)
    backend = emcee.backends.HDFBackend(os.path.join(outdir,fname+'mcmc_save.h5')) # Set up a new backend
    backend.reset(nwalkers, ndim)


    #::: run MCMC
    def run_mcmc(sampler):
        p0 = initial + 1e-8 * np.random.randn(nwalkers, ndim)
        sampler.run_mcmc(p0, total_steps/thin_by, thin_by=thin_by, progress=True);
    
    if multiprocess:    
        with closing(Pool(processes=(multiprocess_cores))) as pool:
            sampler = emcee.EnsembleSampler(nwalkers, ndim, log_probability, pool=pool, backend=backend)
            run_mcmc(sampler)
    else:
        sampler = emcee.EnsembleSampler(nwalkers, ndim, log_probability, backend=backend)
        run_mcmc(sampler)
    
    logprint('\nAcceptance fractions:')
    logprint(sampler.acceptance_fraction)

    tau = sampler.get_autocorr_time(discard=discard, c=5, tol=10, quiet=True)*thin_by
    logprint('\nAutocorrelation times:')
    logprint('\t', '{0: <30}'.format('parameter'), '{0: <20}'.format('tau (in steps)'), '{0: <20}'.format('Chain length (in multiples of tau)'))
    for i, key in enumerate(names):
        logprint('\t', '{0: <30}'.format(key), '{0: <20}'.format(tau[i]), '{0: <20}'.format((total_steps-burn_steps) / tau[i]))
    
    

        
    ###########################################################################
    #::: Output
    ###########################################################################
    def get_params_from_samples(samples, names):
        '''
        read MCMC results and update params
        '''
    
        buf = map(lambda v: (v[1], v[2]-v[1], v[1]-v[0]), zip(*np.percentile(samples, [16, 50, 84], axis=0)))                                         
        theta_median = [ item[0] for item in buf ]
        theta_ul = [ item[1] for item in buf ]
        theta_ll = [ item[2] for item in buf ]
        params_median = { n:t for n,t in zip(names,theta_median) }
        params_ul = { n:t for n,t in zip(names,theta_ul) }
        params_ll = { n:t for n,t in zip(names,theta_ll) }
        
        return params_median, params_ll, params_ul
            
        
    #::: get the samples, 
    samples = sampler.get_chain(flat=True, discard=discard)
    
    
    #::: get the resulting params dictionaries
    params, params_ll, params_ul = get_params_from_samples(samples, names)

    
    #::: Save the resulting parameters in a table
    with open( os.path.join(outdir,fname+'table.csv'), 'wb' ) as f:
        f.write('name,median,ll,ul\n')
        for i, key in enumerate(names):
            f.write(key + ',' + str(params[key]) + ',' + str(params_ll[key]) + ',' + str(params_ul[key]) + '\n' )
    
    
    #::: if requested, append a row into the summary file, too
    if fname_summary is not None:
        with open( fname_summary, 'ab' ) as f:
            f.write(fname[0:-1] + ',')
            for i, key in enumerate(names):
                f.write(str(params[key]) + ',' + str(params_ll[key]) + ',' + str(params_ul[key]))
                if i<len(names)-1:
                    f.write(',')
                else:
                    f.write('\n')


    #::: plot chains; format of chain = (nwalkers, nsteps, nparameters)
    fig, axes = plt.subplots(ndim+1, 1, figsize=(6,4*(ndim+1)) )
    steps = np.arange(0,total_steps,thin_by)
    
    
    #::: plot the lnprob_values (nwalkers, nsteps)
    for j in range(nwalkers):
        axes[0].plot(steps, sampler.get_log_prob()[:,j], '-')
    axes[0].set( ylabel='lnprob', xlabel='steps' )
    
    
    #:::plot all chains of parameters
    for i in range(ndim):
        ax = axes[i+1]
        ax.set( ylabel=names[i], xlabel='steps')
        for j in range(nwalkers):
            ax.plot(steps, sampler.chain[j,:,i], '-')
        ax.axvline( burn_steps, color='k', linestyle='--' )
    
    plt.tight_layout()
    fig.savefig( os.path.join(outdir,fname+'mcmc_chains.jpg'), dpi=100, bbox_inches='tight')
    plt.close(fig)
 
        
    #::: plot corner
    fig = corner.corner(samples,
                        labels=names,
                        show_titles=True, title_kwargs={"fontsize": 12});
    fig.savefig( os.path.join(outdir,fname+'mcmc_corner.jpg'), dpi=100, bbox_inches='tight')
    plt.close(fig)


    logprint('\nDone. All output files are in '+outdir)
    


if __name__ == '__main__':
    pass



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 00:13:48 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: allesfitter modules
from .basement import Basement




def init(datadir, quiet=False):
    global BASEMENT
    BASEMENT = Basement(datadir, quiet=quiet)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov  6 18:32:01 2020

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import json

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def write_csv(fname, *arrays, **kwargs):
    '''
    Writes multiple arrays to a csv file, e.g., time, flux, flux_err.

    Parameters
    ----------
    fname : str
        Name of the file (and directory).
    *arrays : collection of arrays
        One or multiple arrays, e.g. time, flux, flux_err.
    **kwargs : collection of keyword arguments
        Any kwargs for np.savetxt(), e.g., fmt=['%.18e','%.12e','%.12e']

    Returns
    -------
    None.

    Outputs
    -------
    Saves a csv file under the given name.
    '''
    X = np.column_stack(arrays)
    np.savetxt(fname, X, delimiter=',', **kwargs)



def read_csv(fname, skip_header=0):
    '''
    Reads a csv file and unpacks the columns.

    Parameters
    ----------
    fname : str
        Name of the file (and directory).

    Returns
    -------
    Collection of arrays, e.g. time, flux, flux_err.
    '''
    return np.genfromtxt(fname, delimiter=',', comments='#', encoding='utf-8', dtype=float, unpack=True, skip_header=skip_header)



def write_json(fname, dic):
    '''
    Writes something to a json file, e.g. a dictionary.

    Parameters
    ----------
    fname : str
        Name of the file (and directory).
    dic : dictionary
        A dictionary (or other collection) to be saved.

    Returns
    -------
    None.

    Outputs
    -------
    Saves a json file under the given name.
    '''
    with open(fname, 'w') as fp:
        json.dump(dic, fp, indent=4)



def read_json(fname):
    '''
    Reads a json file and retrieves the content.

    Parameters
    ----------
    fname : str
        Name of the file (and directory).

    Returns
    -------
    A dictionary (or other collection).
    '''
    with open(fname, 'r') as fp:
        dic = json.load(fp)
    return dic


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 00:17:06 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import os
import sys
import fnmatch
import collections
from datetime import datetime
from multiprocessing import cpu_count
import warnings
warnings.formatwarning = lambda msg, *args, **kwargs: f'\n! WARNING:\n {msg}\ntype: {args[0]}, file: {args[1]}, line: {args[2]}\n'
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning) 
warnings.filterwarnings('ignore', category=np.RankWarning) 
from scipy.stats import truncnorm

#::: allesfitter modules
from .exoworlds_rdx.lightcurves.index_transits import index_transits, index_eclipses, get_first_epoch, get_tmid_observed_transits
from .priors.simulate_PDF import simulate_PDF
from .utils.mcmc_move_translator import translate_str_to_move

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})
                     
    
    
    
###############################################################################
#::: 'Basement' class, which contains all the data, settings, etc.
###############################################################################
class Basement():
    '''
    The 'Basement' class contains all the data, settings, etc.
    '''
    
    ###############################################################################
    #::: init
    ###############################################################################
    def __init__(self, datadir, quiet=False):
        '''
        Inputs:
        -------
        datadir : str
            the working directory for allesfitter
            must contain all the data files
            output directories and files will also be created inside datadir
        fast_fit : bool (optional; default is False)
            if False: 
                use all photometric data for the plot
            if True: 
                only use photometric data in an 8h window around the transit 
                requires a good initial guess of the epoch and period
                
        Returns:
        --------
        All the variables needed for allesfitter
        '''
        
        print('Filling the Basement')
        
        self.quiet = quiet
        self.now = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.now())
        self.datadir = datadir
        self.outdir = os.path.join(datadir,'results') 
        if not os.path.exists( self.outdir ): os.makedirs( self.outdir )
        
        print('')
        self.logprint('\nallesfitter version')
        self.logprint('---------------------')
        self.logprint('v1.2.9')
        
        self.load_settings()
        self.load_params()
        self.load_data()
        
        if self.settings['shift_epoch']:
            try:
                self.change_epoch()
            except:
                warnings.warn('\nCould not shift epoch (you can peacefully ignore this warning if no period was given)\n')
        
        if self.settings['fit_ttvs']:  
            self.setup_ttv_fit()
        
        #::: external priors (e.g. stellar density)
        self.external_priors = {}
        self.load_stellar_priors()
        
        #::: if baseline model == sample_GP, set up a GP object for photometric data
#        self.setup_GPs()
        
        #::: translate limb darkening codes from params.csv (int) into str for ellc
        self.ldcode_to_ldstr = ["none",#   :  0,
                                "lin",#    :  1,
                                "quad",#   :  2,
                                "sing",#   :  3,
                                "claret",# :  4,
                                "log",#  :  5,
                                "sqrt",#  :  6,
                                "exp",#    :  7,
                                "power-2",#:  8,
                                "mugrid"]# : -1

        #::: check if the input is consistent
        for inst in self.settings['inst_phot']:
            key='flux'
            if (self.settings['baseline_'+key+'_'+inst] in ['sample_GP_Matern32', 'sample_GP_SHO']) &\
               (self.settings['error_'+key+'_'+inst] != 'sample'):
                   raise ValueError('If you want to use '+self.settings['baseline_'+key+'_'+inst]+', you will want to sample the jitters, too!')
            
                 
                    
    ###############################################################################
    #::: print function that prints into console and logfile at the same time
    ############################################################################### 
    def logprint(self, *text):
        if not self.quiet:
            print(*text)
            original = sys.stdout
            with open( os.path.join(self.outdir,'logfile_'+self.now+'.log'), 'a' ) as f:
                sys.stdout = f
                print(*text)
            sys.stdout = original
        else:
            pass
        
        

    ###############################################################################
    #::: load settings
    ###############################################################################
    def load_settings(self):
        '''
        For the full list of options see www.allesfitter.com
        '''
        
        
        def set_bool(text):
            if text.lower() in ['true', '1']:
                return True
            else:
                return False
            
        
        def is_empty_or_none(key):
            return (key not in self.settings) or (str(self.settings[key]).lower() == 'none') or (len(self.settings[key])==0)
            
        
        def unique(array):
            uniq, index = np.unique(array, return_index=True)
            return uniq[index.argsort()]
            
        
        rows = np.genfromtxt( os.path.join(self.datadir,'settings.csv'),dtype=None,encoding='utf-8',delimiter=',' )

        #::: make backwards compatible
        for i, row in enumerate(rows):
#            print(row)
            name = row[0]
            if name[:7]=='planets':
                rows[i][0] = 'companions'+name[7:]
                warnings.warn('You are using outdated keywords. Automatically renaming '+name+' ---> '+rows[i][0]+'. Please fix this before the Duolingo owl comes to get you.') #, category=DeprecationWarning)
            if name[:6]=='ld_law':
                rows[i][0] = 'host_ld_law'+name[6:]
                warnings.warn('You are using outdated keywords. Automatically renaming '+name+' ---> '+rows[i][0]+'. Please fix this before the Duolingo owl comes to get you.') #, category=DeprecationWarning)
                
#        self.settings = {r[0]:r[1] for r in rows}
        self.settings = collections.OrderedDict( [('user-given:','')]+[ (r[0],r[1] ) for r in rows ]+[('automatically set:','')] )

        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Main settings
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'time_format' not in self.settings:
            self.settings['time_format'] = 'BJD_TDB'
            
        for key in ['companions_phot', 'companions_rv', 'inst_phot', 'inst_rv', 'inst_rv2']:
            if key not in self.settings:
                self.settings[key] = []
            elif len(self.settings[key]): 
                self.settings[key] = str(self.settings[key]).split(' ')
            else:                       
                self.settings[key] = []
        
        self.settings['companions_all']  = list(np.unique(self.settings['companions_phot']+self.settings['companions_rv'])) #sorted by b, c, d...
        self.settings['inst_all'] = list(unique( self.settings['inst_phot']+self.settings['inst_rv']+self.settings['inst_rv2'] )) #sorted like user input
    
        if len(self.settings['inst_phot'])==0 and len(self.settings['companions_phot'])>0:
            raise ValueError('No photometric instrument is selected, but photometric companions are given.')
        if len(self.settings['inst_rv'])==0 and len(self.settings['companions_rv'])>0:
           raise ValueError('No RV instrument is selected, but RV companions are given.')
            
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: General settings
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'print_progress' in self.settings:
            self.settings['print_progress'] = set_bool(self.settings['print_progress'] )
        else:
            self.settings['print_progress'] = True
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Epoch settings
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'shift_epoch' in self.settings:
            self.settings['shift_epoch'] = set_bool(self.settings['shift_epoch'] )
        else:
            self.settings['shift_epoch'] = True 
            
            
        for companion in self.settings['companions_all']:
            if 'inst_for_'+companion+'_epoch' not in self.settings:
                self.settings['inst_for_'+companion+'_epoch'] = 'all'
        
            if self.settings['inst_for_'+companion+'_epoch'] in ['all','none']:
                self.settings['inst_for_'+companion+'_epoch'] = self.settings['inst_all']
            else:
                if len(self.settings['inst_for_'+companion+'_epoch']): 
                    self.settings['inst_for_'+companion+'_epoch'] = str(self.settings['inst_for_'+companion+'_epoch']).split(' ')
                else:                       
                    self.settings['inst_for_'+companion+'_epoch'] = []
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Multiprocess settings
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        self.settings['multiprocess'] = set_bool(self.settings['multiprocess'])
        
        from pprint import pprint
        pprint(self.settings)
        
        if 'multiprocess_cores' not in self.settings.keys():
            self.settings['multiprocess_cores'] = cpu_count()-1
        elif self.settings['multiprocess_cores'] == 'all':
            self.settings['multiprocess_cores'] = cpu_count()-1
        else:
            self.settings['multiprocess_cores'] = int(self.settings['multiprocess_cores'])
            if self.settings['multiprocess_cores'] == cpu_count():
                string = 'You are pushing your luck: you want to run on '+str(self.settings['multiprocess_cores'])+' cores, but your computer has only '+str(cpu_count())+'. I will let you go through with it this time...'
                warnings.warn(string)
            if self.settings['multiprocess_cores'] > cpu_count():
                string = 'Oops, you want to run on '+str(self.settings['multiprocess_cores'])+' cores, but your computer has only '+str(cpu_count())+'. Maybe try running on '+str(cpu_count()-1)+'?'


        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Phase variations
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if ('phase_variations' in self.settings.keys()) and len(self.settings['phase_variations']):
            warnings.warn('You are using outdated keywords. Automatically renaming "phase_variations" ---> "phase_curve".'+'. Please fix this before the Duolingo owl comes to get you.')
            self.settings['phase_curve'] = self.settings['phase_variations']
            
        if ('phase_curve' in self.settings.keys()) and len(self.settings['phase_curve']):
            self.settings['phase_curve'] = set_bool(self.settings['phase_curve'])
            if self.settings['phase_curve']==True:                
                # self.logprint('The user set phase_curve==True. Automatically set fast_fit=False and secondary_eclispe=True, and overwrite other settings.')
                self.settings['fast_fit'] = 'False'
                self.settings['secondary_eclipse'] = 'True'
        else:
            self.settings['phase_curve'] = False
            
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Fast fit
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if ('fast_fit' in self.settings.keys()) and len(self.settings['fast_fit']):
            self.settings['fast_fit'] = set_bool(self.settings['fast_fit'])
        else:
            self.settings['fast_fit'] = False
        
        if ('fast_fit_width' in self.settings.keys()) and len(self.settings['fast_fit_width']):
            self.settings['fast_fit_width'] = np.float(self.settings['fast_fit_width'])
        else:
            self.settings['fast_fit_width'] = 8./24.
                
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Host stellar density prior
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'use_host_density_prior' in self.settings:
            self.settings['use_host_density_prior'] = set_bool(self.settings['use_host_density_prior'] )
        else:
            self.settings['use_host_density_prior'] = True
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Host stellar density prior
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'use_tidal_eccentricity_prior' in self.settings:
            self.settings['use_tidal_eccentricity_prior'] = set_bool(self.settings['use_tidal_eccentricity_prior'] )
        else:
            self.settings['use_tidal_eccentricity_prior'] = False
            
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: TTVs
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if ('fit_ttvs' in self.settings.keys()) and len(self.settings['fit_ttvs']):
            self.settings['fit_ttvs'] = set_bool(self.settings['fit_ttvs'])
            if (self.settings['fit_ttvs']==True) and (self.settings['fast_fit']==False):
                raise ValueError('fit_ttvs==True, but fast_fit==False.'+\
                                 'Currently, you can only fit for TTVs if fast_fit==True.'+\
                                 'Please choose different settings.')
        else:
            self.settings['fit_ttvs'] = False
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Secondary eclipse
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if ('secondary_eclipse' in self.settings.keys()) and len(self.settings['secondary_eclipse']):
            self.settings['secondary_eclipse'] = set_bool(self.settings['secondary_eclipse'])
        else:
            self.settings['secondary_eclipse'] = False
                        
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: MCMC settings
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'mcmc_pre_run_loops' not in self.settings: 
            self.settings['mcmc_pre_run_loops'] = 0
        if 'mcmc_pre_run_steps' not in self.settings: 
            self.settings['mcmc_pre_run_steps'] = 0
        if 'mcmc_nwalkers' not in self.settings: 
            self.settings['mcmc_nwalkers'] = 100
        if 'mcmc_total_steps' not in self.settings: 
            self.settings['mcmc_total_steps'] = 2000
        if 'mcmc_burn_steps' not in self.settings: 
            self.settings['mcmc_burn_steps'] = 1000
        if 'mcmc_thin_by' not in self.settings: 
            self.settings['mcmc_thin_by'] = 1
        if 'mcmc_moves' not in self.settings: 
            self.settings['mcmc_moves'] = 'DEMove'
            
        #::: make sure these are integers
        for key in ['mcmc_nwalkers','mcmc_pre_run_loops','mcmc_pre_run_steps',
                    'mcmc_total_steps','mcmc_burn_steps','mcmc_thin_by']:
            self.settings[key] = int(self.settings[key])
            
        #::: luser proof
        if self.settings['mcmc_total_steps'] <= self.settings['mcmc_burn_steps']:
            raise ValueError('Your setting for mcmc_total_steps must be larger than mcmc_burn_steps (check your settings.csv).')
                
            
        #::: translate the mcmc_move string into a list of emcee commands
        self.settings['mcmc_moves'] = translate_str_to_move(self.settings['mcmc_moves'])
            
        # N_evaluation_samples = int( 1. * self.settings['mcmc_nwalkers'] * (self.settings['mcmc_total_steps']-self.settings['mcmc_burn_steps']) / self.settings['mcmc_thin_by'] )
        # self.logprint('\nAnticipating ' + str(N_evaluation_samples) + 'MCMC evaluation samples.\n')
        # if N_evaluation_samples>200000:
        #     answer = input('It seems like you are asking for ' + str(N_evaluation_samples) + 'MCMC evaluation samples (calculated as mcmc_nwalkers * (mcmc_total_steps-mcmc_burn_steps) / mcmc_thin_by).'+\
        #                    'That is an aweful lot of samples.'+\
        #                    'What do you want to do?\n'+\
        #                    '1 : continue at any sacrifice\n'+\
        #                    '2 : abort and increase the mcmc_thin_by parameter in settings.csv (do not do this if you continued an old run!)\n')
        #     if answer==1: 
        #         pass
        #     else:
        #         raise ValueError('User aborted the run.')

        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Nested Sampling settings
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'ns_modus' not in self.settings: 
            self.settings['ns_modus'] = 'static'
        if 'ns_nlive' not in self.settings: 
            self.settings['ns_nlive'] = 500
        if 'ns_bound' not in self.settings: 
            self.settings['ns_bound'] = 'single'
        if 'ns_sample' not in self.settings: 
            self.settings['ns_sample'] = 'rwalk'
        if 'ns_tol' not in self.settings: 
            self.settings['ns_tol'] = 0.01
                
        self.settings['ns_nlive'] = int(self.settings['ns_nlive'])
        self.settings['ns_tol'] = float(self.settings['ns_tol'])
        
#        if self.settings['ns_sample'] == 'auto':
#            if self.ndim < 10:
#                self.settings['ns_sample'] = 'unif'
#                print('Using ns_sample=="unif".')
#            elif 10 <= self.ndim <= 20:
#                self.settings['ns_sample'] = 'rwalk'
#                print('Using ns_sample=="rwalk".')
#            else:
#                self.settings['ns_sample'] = 'slice'
#                print('Using ns_sample=="slice".')
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: host & companion grids, limb darkening laws, shapes, etc.
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for companion in self.settings['companions_all']:
            for inst in self.settings['inst_all']:
                
                if 'host_grid_'+inst not in self.settings: 
                    self.settings['host_grid_'+inst] = 'default'
                    
                if companion+'_grid_'+inst not in self.settings: 
                    self.settings[companion+'_grid_'+inst] = 'default'
                    
                if is_empty_or_none('host_ld_law_'+inst): 
                    self.settings['host_ld_law_'+inst] = None
                    
                if is_empty_or_none(companion+'_ld_law_'+inst):
                    self.settings[companion+'_ld_law_'+inst] = None
 
                if is_empty_or_none('host_ld_space_'+inst): 
                    self.settings['host_ld_space_'+inst] = 'q'
                    
                if is_empty_or_none(companion+'_ld_space_'+inst):
                    self.settings[companion+'_ld_space_'+inst] = 'q'        
                    
                if 'host_shape_'+inst not in self.settings: 
                    self.settings['host_shape_'+inst] = 'sphere'
                    
                if companion+'_shape_'+inst not in self.settings: 
                    self.settings[companion+'_shape_'+inst] = 'sphere'
                    
                    
        for companion in self.settings['companions_rv']:
            for inst in list(self.settings['inst_rv']) + list(self.settings['inst_rv2']):
                if companion+'_flux_weighted_'+inst in self.settings: 
                    self.settings[companion+'_flux_weighted_'+inst] = set_bool(self.settings[companion+'_flux_weighted_'+inst])
                else:
                    self.settings[companion+'_flux_weighted_'+inst] = False
        
    
        if 'exact_grav' in self.settings: 
            self.settings['exact_grav'] = set_bool(self.settings['exact_grav'])
        else:
            self.settings['exact_grav'] = False
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Phase curve styles
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if is_empty_or_none('phase_curve_style'):
            self.settings['phase_curve_style'] = None
        if self.settings['phase_curve_style'] not in [None, 'sine_series', 'sine_physical', 'ellc_physical', 'GP']:
            raise ValueError("The setting 'phase_curve_style' must be one of [None, 'sine_series', 'sine_physical', 'ellc_physical', 'GP'], but was '"+str(self.settings['phase_curve_style'])+"'.")
        if (self.settings['phase_curve'] is True) and (self.settings['phase_curve_style'] is None):
            raise ValueError("You chose 'phase_curve=True' but did not select a 'phase_curve_style'; please select one of ['sine_series', 'sine_physical', 'ellc_physical', 'GP'].")
        if (self.settings['phase_curve'] is False) and (self.settings['phase_curve_style'] in ['sine_series', 'sine_physical', 'ellc_physical', 'GP']):
           raise ValueError("You chose 'phase_curve=False' but also selected a 'phase_curve_style'; please double check and set 'phase_curve_style=None' (or remove it).")
                               
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Stellar variability
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for key in ['flux', 'rv', 'rv2']:
            if ('stellar_var_'+key not in self.settings) or (self.settings['stellar_var_'+key] is None) or (self.settings['stellar_var_'+key].lower()=='none'): 
                self.settings['stellar_var_'+key] = 'none'
                     
                     
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Baselines
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for inst in self.settings['inst_all']:
            if inst in self.settings['inst_phot']: key='flux'
            elif inst in self.settings['inst_rv']: key='rv'
            elif inst in self.settings['inst_rv2']: key='rv2'
            
            #::: default
            #::: if the user gives no baseline, the default is 'none'
            if 'baseline_'+key+'_'+inst not in self.settings: 
                self.settings['baseline_'+key+'_'+inst] = 'none'

            #::: hybrid_spline
            #::: the user can define the s value directly, e.g. as "hybrid_spline 0.001"
            #::: this block serves to split up this input and assign it to the right functions
            if ('hybrid_spline' in self.settings['baseline_'+key+'_'+inst])\
                and (len(self.settings['baseline_'+key+'_'+inst].split(' '))>1): 
                s = self.settings['baseline_'+key+'_'+inst].split(' ')[1]
                self.settings['baseline_'+key+'_'+inst] = 'hybrid_spline_s'
                self.settings['baseline_'+key+'_'+inst+'_args'] = s #any arguments coming with this baseline (for future expandability; for now it is simply the s-value)
                
            #::: sample_GP
            #::: make sure the keywords are updated correctly
            elif self.settings['baseline_'+key+'_'+inst] == 'sample_GP': 
                 warnings.warn('You are using outdated keywords. Automatically renaming sample_GP ---> sample_GP_Matern32.'+'. Please update your files before the Duolingo owl comes to get you.') #, category=DeprecationWarning)
                 self.settings['baseline_'+key+'_'+inst] = 'sample_GP_Matern32'
                 
            #::: baseline against custom series
            #::: allows the user to fit a baseline not vs. time but vs. a chosen custom series
            if 'baseline_'+key+'_'+inst+'_against' not in self.settings:
                self.settings['baseline_'+key+'_'+inst+'_against'] = 'time'
            if self.settings['baseline_'+key+'_'+inst+'_against'] not in ['time','custom_series']:
                raise ValueError("The setting 'baseline_'+key+'_'+inst+'_against' must be one of ['time', custom_series'], but was '" + self.settings['baseline_'+key+'_'+inst+'_against'] + "'.")
                     
                
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Errors
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for inst in self.settings['inst_all']:
            if inst in self.settings['inst_phot']: key='flux'
            elif inst in self.settings['inst_rv']: key='rv'
            elif inst in self.settings['inst_rv2']: key='rv2'
            if 'error_'+key+'_'+inst not in self.settings: 
                self.settings['error_'+key+'_'+inst] = 'sample'
            
        # for inst in self.settings['inst_phot']:
        #     for key in ['flux']:
        #         if 'error_'+key+'_'+inst not in self.settings: 
        #             self.settings['error_'+key+'_'+inst] = 'sample'

        # for inst in self.settings['inst_rv']:
        #     for key in ['rv']:
        #         if 'error_'+key+'_'+inst not in self.settings: 
        #             self.settings['error_'+key+'_'+inst] = 'sample'
                    
                
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Color plot
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'color_plot' not in self.settings.keys():
            self.settings['color_plot'] = False
            
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Companion colors
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for i, companion in enumerate( self.settings['companions_all'] ):
            self.settings[companion+'_color'] = sns.color_palette()[i]
        
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Plot zoom window
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'zoom_window' not in self.settings:
            self.settings['zoom_window'] = 8./24. #8h window around transit/eclipse midpoint by Default
        else:
            self.settings['zoom_window'] = float(self.settings['zoom_window'])
            
            
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Exposure time interpolation
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for inst in self.settings['inst_all']:
            #::: if t_exp is given
            if 't_exp_'+inst in self.settings.keys() and len(self.settings['t_exp_'+inst]):
                t_exp = self.settings['t_exp_'+inst].split(' ')
                #if float
                if len(t_exp)==1:
                    self.settings['t_exp_'+inst] = np.float(t_exp[0])
                #if array
                else:
                    self.settings['t_exp_'+inst] = np.array([ np.float(t) for t in t_exp ])
            #::: if not given / given as an empty field
            else:
                self.settings['t_exp_'+inst] = None
                
            #::: if t_exp_n_int is given
            if 't_exp_'+inst in self.settings \
                and 't_exp_n_int_'+inst in self.settings \
                and len(self.settings['t_exp_n_int_'+inst]):
                    
                self.settings['t_exp_n_int_'+inst] = int(self.settings['t_exp_n_int_'+inst])
                if self.settings['t_exp_n_int_'+inst] < 1:
                    raise ValueError('"t_exp_n_int_'+inst+'" must be >= 1, but is given as '+str(self.settings['t_exp_n_int_'+inst])+' in params.csv')
            else:
                self.settings['t_exp_n_int_'+inst] = None
  
    
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Number of spots
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        for inst in self.settings['inst_all']:
            if 'host_N_spots_'+inst in self.settings and len(self.settings['host_N_spots_'+inst]):
                self.settings['host_N_spots_'+inst] = int(self.settings['host_N_spots_'+inst])
            else:
                self.settings['host_N_spots_'+inst] = 0
        
            for companion in self.settings['companions_all']:
                if companion+'_N_spots'+inst in self.settings:
                    self.settings[companion+'_N_spots_'+inst] = int(self.settings[companion+'_N_spots_'+inst])
                else:
                    self.settings[companion+'_N_spots_'+inst] = 0
                    
        
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        #::: Number of flares
        #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        if 'N_flares' in self.settings and len(self.settings['N_flares'])>0:
            self.settings['N_flares'] = int(self.settings['N_flares'])
        else:
            self.settings['N_flares'] = 0
        
        
        
        
                
    ###############################################################################
    #::: load params
    ###############################################################################
    def load_params(self):
        '''
        For the full list of options see www.allesfitter.com
        '''
    
        #==========================================================================
        #::: load params.csv
        #==========================================================================   
        buf = np.genfromtxt(os.path.join(self.datadir,'params.csv'), delimiter=',',comments='#',dtype=None,encoding='utf-8',names=True)
            
           
        #==========================================================================
        #::: function to assure backwards compability
        #==========================================================================
        def backwards_compability(key_new, key_deprecated):
            if key_deprecated in np.atleast_1d(buf['name']):
                warnings.warn('You are using outdated keywords. Automatically renaming '+key_deprecated+' ---> '+key_new+'. Please fix this before the Duolingo owl comes to get you.') #, category=DeprecationWarning)
                ind = np.where(buf['name'] == key_deprecated)[0]
                np.atleast_1d(buf['name'])[ind] = key_new
                
                
        #==========================================================================
        #::: luser-proof: backwards compability 
        # (has to happend first thing and right inside buf['name'])
        #==========================================================================
        for inst in self.settings['inst_all']:
            backwards_compability(key_new='host_ldc_q1_'+inst, key_deprecated='ldc_q1_'+inst)
            backwards_compability(key_new='host_ldc_q2_'+inst, key_deprecated='ldc_q2_'+inst)
            backwards_compability(key_new='host_ldc_q3_'+inst, key_deprecated='ldc_q3_'+inst)
            backwards_compability(key_new='host_ldc_q4_'+inst, key_deprecated='ldc_q4_'+inst)
            backwards_compability(key_new='ln_err_flux_'+inst, key_deprecated='log_err_flux_'+inst)
            backwards_compability(key_new='ln_jitter_rv_'+inst, key_deprecated='log_jitter_rv_'+inst)
            backwards_compability(key_new='baseline_gp_matern32_lnsigma_flux_'+inst, key_deprecated='baseline_gp1_flux_'+inst)
            backwards_compability(key_new='baseline_gp_matern32_lnrho_flux_'+inst, key_deprecated='baseline_gp2_flux_'+inst)
            backwards_compability(key_new='baseline_gp_matern32_lnsigma_rv_'+inst, key_deprecated='baseline_gp1_rv_'+inst)
            backwards_compability(key_new='baseline_gp_matern32_lnrho_rv_'+inst, key_deprecated='baseline_gp2_rv_'+inst)
                   
                    
        #==========================================================================
        #::: luser-proof: check for allowed keys to catch typos etc.
        #==========================================================================  
        #TODO
                
                
        #==========================================================================
        #::: set up stuff   
        #==========================================================================          
        self.allkeys = np.atleast_1d(buf['name']) #len(all rows in params.csv)
        self.labels = np.atleast_1d(buf['label']) #len(all rows in params.csv)
        self.units = np.atleast_1d(buf['unit']) #len(all rows in params.csv)
        if 'truth' in buf.dtype.names:
            self.truths = np.atleast_1d(buf['truth']) #len(all rows in params.csv)
        else:
            self.truths = np.nan * np.ones(len(self.allkeys))
            
        self.params = collections.OrderedDict() #len(all rows in params.csv)
        self.params['user-given:'] = '' #just for pretty printing
        for i,key in enumerate(self.allkeys):
            #::: if it's not a "coupled parameter", then use the given value
            if np.atleast_1d(buf['value'])[i] not in list(self.allkeys):
                self.params[key] = np.float(np.atleast_1d(buf['value'])[i])
            #::: if it's a "coupled parameter", then write the string of the key it is coupled to
            else:
                self.params[key] = np.atleast_1d(buf['value'])[i]
                
                
        #==========================================================================
        #::: function to automatically set default params if they were not given
        #==========================================================================
        def validate(key, default, default_min, default_max):
            if (key in self.params) and (self.params[key] is not None):
                if (self.params[key] < default_min) or (self.params[key] > default_max):
                    raise ValueError("User input for "+key+" is "+self.params+" but must lie within ["+str(default_min)+","+str(default_max)+"].")
            if (key not in self.params):
                self.params[key] = default
        
        
        #==========================================================================
        #::: luser-proof: make sure the limb darkening values are uniquely 
        #::: from either the u- or q-space
        #==========================================================================  
        def check_ld(obj, inst):
           if self.settings[obj+'_ld_space_'+inst] == 'q': 
                matches = fnmatch.filter(self.allkeys, obj+'_ldc_u*_'+inst)
                if len(matches) > 0:
                    raise ValueError("The following user input is inconsistent:\n"+\
                                     "Setting: '"+key+"' = 'q'\n"+\
                                     "Parameters: {}".format(matches))   
                        
           elif self.settings[obj+'_ld_space_'+inst] == 'u': 
                matches = fnmatch.filter(self.allkeys, obj+'_ldc_q*_'+inst)
                if len(matches) > 0:
                    raise ValueError("The following user input is inconsistent:\n"+\
                                     "Setting: '"+key+"' = 'u'\n"+\
                                     "Parameters: {}".format(matches))  
                        
        for inst in self.settings['inst_all']:
            for obj in ['host'] + self.settings['companions_all']:   
                check_ld(obj, inst)
            
            
        #==========================================================================
        #::: validate that initial guess params have reasonable values
        #==========================================================================
        self.params['automatically set:'] = '' #just for pretty printing
        for companion in self.settings['companions_all']:
            for inst in self.settings['inst_all']:
                
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: ellc defaults
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                
                #::: frequently used parameters
                validate(companion+'_rr', None, 0., np.inf)
                validate(companion+'_rsuma', None, 0., np.inf)
                validate(companion+'_cosi', 0., 0., 1.)
                validate(companion+'_epoch', 0., -np.inf, np.inf)
                validate(companion+'_period', 0., 0., np.inf)
                validate(companion+'_sbratio_'+inst, 0., 0., np.inf)
                validate(companion+'_K', 0., 0., np.inf)
                validate(companion+'_f_s', 0., -1, 1)
                validate(companion+'_f_c', 0., -1, 1)
                validate('dil_'+inst, 0., -np.inf, np.inf)
                
                #::: limb darkenings, u-space
                validate('host_ldc_u1_'+inst, None, 0, 1)
                validate('host_ldc_u2_'+inst, None, 0, 1)
                validate('host_ldc_u3_'+inst, None, 0, 1)
                validate('host_ldc_u4_'+inst, None, 0, 1)
                validate(companion+'_ldc_u1_'+inst, None, 0, 1)
                validate(companion+'_ldc_u2_'+inst, None, 0, 1)
                validate(companion+'_ldc_u3_'+inst, None, 0, 1)
                validate(companion+'_ldc_u4_'+inst, None, 0, 1)

                #::: limb darkenings, q-space
                validate('host_ldc_q1_'+inst, None, 0, 1)
                validate('host_ldc_q2_'+inst, None, 0, 1)
                validate('host_ldc_q3_'+inst, None, 0, 1)
                validate('host_ldc_q4_'+inst, None, 0, 1)
                validate(companion+'_ldc_q1_'+inst, None, 0, 1)
                validate(companion+'_ldc_q2_'+inst, None, 0, 1)
                validate(companion+'_ldc_q3_'+inst, None, 0, 1)
                validate(companion+'_ldc_q4_'+inst, None, 0, 1)
                
                #::: catch exceptions
                if self.params[companion+'_period'] is None:
                    self.settings['do_not_phase_fold'] = True
                
                #::: advanced parameters
                validate(companion+'_a', None, 0., np.inf)
                validate(companion+'_q', 1., 0., np.inf)
                
                validate('didt_'+inst, None, -np.inf, np.inf)
                validate('domdt_'+inst, None, -np.inf, np.inf)
                
                validate('host_gdc_'+inst, None, 0., 1.)
                validate('host_rotfac_'+inst, 1., 0., np.inf)
                validate('host_hf_'+inst, 1.5, -np.inf, np.inf)
                validate('host_bfac_'+inst, None, -np.inf, np.inf)
                validate('host_heat_'+inst, None, -np.inf, np.inf)
                validate('host_lambda', None, -np.inf, np.inf)
                validate('host_vsini', None, -np.inf, np.inf)
                
                validate(companion+'_gdc_'+inst, None, 0., 1.)
                validate(companion+'_rotfac_'+inst, 1., 0., np.inf)
                validate(companion+'_hf_'+inst, 1.5, -np.inf, np.inf)
                validate(companion+'_bfac_'+inst, None, -np.inf, np.inf)
                validate(companion+'_heat_'+inst, None, -np.inf, np.inf)
                validate(companion+'_lambda', None, -np.inf, np.inf)
                validate(companion+'_vsini', None, -np.inf, np.inf)
        
                #::: special parameters (list type)
                if 'host_spots_'+inst not in self.params:
                    self.params['host_spots_'+inst] = None
                if companion+'_spots_'+inst not in self.params:
                    self.params[companion+'_spots_'+inst] = None
                    
                
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: errors and jitters
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #TODO: add validations for all errors / jitters
                    
                
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: baselines (and backwards compability)
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #TODO: add validations for all baseline params
                
                    
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: phase curve style: sine_series
                # all in ppt
                # A1 (beaming)
                # B1 (atmospheric), can be split in thermal and reflected
                # B2 (ellipsoidal)
                # B3 (ellipsoidal 2nd order)
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                # if (self.settings['phase_curve_style'] == 'sine_series') and (inst in self.settings['inst_phot']):
                if (inst in self.settings['inst_phot']):
                    validate(companion+'_phase_curve_A1_'+inst, None, 0., np.inf)
                    validate(companion+'_phase_curve_B1_'+inst, None, -np.inf, 0.)
                    validate(companion+'_phase_curve_B1_shift_'+inst, 0., -np.inf, np.inf)
                    validate(companion+'_phase_curve_B1t_'+inst, None, -np.inf, 0.)
                    validate(companion+'_phase_curve_B1t_shift_'+inst, 0., -np.inf, np.inf)
                    validate(companion+'_phase_curve_B1r_'+inst, None, -np.inf, 0.)
                    validate(companion+'_phase_curve_B1r_shift_'+inst, 0., -np.inf, np.inf)
                    validate(companion+'_phase_curve_B2_'+inst, None, -np.inf, 0.)
                    validate(companion+'_phase_curve_B3_'+inst, None, -np.inf, 0.)

                                       
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: phase curve style: sine_physical  
                # A1 (beaming)
                # B1 (atmospheric), can be split in thermal and reflected
                # B2 (ellipsoidal)
                # B3 (ellipsoidal 2nd order)  
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                # if (self.settings['phase_curve_style'] == 'sine_physical') and (inst in self.settings['inst_phot']):
                if (inst in self.settings['inst_phot']):
                    validate(companion+'_phase_curve_beaming_'+inst, None, 0., np.inf)
                    validate(companion+'_phase_curve_atmospheric_'+inst, None, 0., np.inf)
                    validate(companion+'_phase_curve_atmospheric_shift_'+inst, 0., -np.inf, np.inf)
                    validate(companion+'_phase_curve_atmospheric_thermal_'+inst, None, 0., np.inf)
                    validate(companion+'_phase_curve_atmospheric_thermal_shift_'+inst, 0., -np.inf, np.inf)
                    validate(companion+'_phase_curve_atmospheric_reflected_'+inst, None, 0., np.inf)
                    validate(companion+'_phase_curve_atmospheric_reflected_shift_'+inst, 0., -np.inf, np.inf)
                    validate(companion+'_phase_curve_ellipsoidal_'+inst, None, 0., np.inf)
                    validate(companion+'_phase_curve_ellipsoidal_2nd_'+inst, None, 0., np.inf)
                
                
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: to avoid a bug/feature in ellc, if either property is >0, set the other to 1-15 (not 0):
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                if self.params[companion+'_heat_'+inst] is not None:
                    if (self.params[companion+'_sbratio_'+inst] == 0) and (self.params[companion+'_heat_'+inst] > 0):
                        self.params[companion+'_sbratio_'+inst] = 1e-15        #this is to avoid a bug/feature in ellc
                    if (self.params[companion+'_sbratio_'+inst] > 0) and (self.params[companion+'_heat_'+inst] == 0):
                        self.params[companion+'_heat_'+inst] = 1e-15           #this is to avoid a bug/feature in ellc
              

                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                #::: luser proof: avoid conflicting/degenerate phase curve commands
                #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
                if (inst in self.settings['inst_phot']) and (self.settings['phase_curve'] == True):
                    phase_curve_model_1 = (self.params[companion+'_phase_curve_B1_'+inst] is not None)
                    phase_curve_model_2 = ((self.params[companion+'_phase_curve_B1t_'+inst] is not None) or (self.params[companion+'_phase_curve_B1r_'+inst] is not None))
                    phase_curve_model_3 = (self.params[companion+'_phase_curve_atmospheric_'+inst] is not None)
                    phase_curve_model_4 = ((self.params[companion+'_phase_curve_atmospheric_thermal_'+inst] is not None) or (self.params[companion+'_phase_curve_atmospheric_reflected_'+inst] is not None))
                    phase_curve_model_5 = ((self.params['host_bfac_'+inst] is not None) or (self.params['host_heat_'+inst] is not None) or \
                                           (self.params['host_gdc_'+inst] is not None) or (self.settings['host_shape_'+inst]!='sphere') or \
                                           (self.params[companion+'_bfac_'+inst] is not None) or (self.params[companion+'_heat_'+inst] is not None) or \
                                           (self.params[companion+'_gdc_'+inst] is not None) or (self.settings[companion+'_shape_'+inst]!='sphere'))
                    if (phase_curve_model_1 + phase_curve_model_2 + phase_curve_model_3 + phase_curve_model_4 + phase_curve_model_5) > 1:
                        raise ValueError('You can use either\n'\
                                         +'1) the sine_series phase curve model with "*_phase_curve_B1_*",\n'\
                                         +'2) the sine_series phase curve model with "*_phase_curve_B1t_*" and "*_phase_curve_B1r_*", or\n'\
                                         +'3) the sine_physical phase curve model with "*_phase_curve_atmospheric_*",\n'\
                                         +'4) the sine_physical phase curve model with "*_phase_curve_atmospheric_thermal_*" and "*_phase_curve_atmospheric_reflected_*", or\n'\
                                         +'5) the ellc_physical phase curve model with "*_bfac_*", "*_heat_*", "*_gdc_*" etc.\n'\
                                         +'but you shall not pass with a mix&match.')
                    
                        
        #==========================================================================
        #::: coupled params
        #==========================================================================
        if 'coupled_with' in buf.dtype.names:
            self.coupled_with = buf['coupled_with']
        else:
            self.coupled_with = [None]*len(self.allkeys)
            
        for i, key in enumerate(self.allkeys):
            if isinstance(self.coupled_with[i], str) and (len(self.coupled_with[i])>0):
                self.params[key] = self.params[self.coupled_with[i]]           #luser proof: automatically set the values of the params coupled to another param
                buf['fit'][i] = 0                                              #luser proof: automatically set fit=0 for the params coupled to another param
        
        
        #==========================================================================
        #::: mark to be fitted params
        #==========================================================================
        self.ind_fit = (buf['fit']==1)                  #len(all rows in params.csv)
        
        self.fitkeys = buf['name'][ self.ind_fit ]      #len(ndim)
        self.fitlabels = self.labels[ self.ind_fit ]    #len(ndim)
        self.fitunits = self.units[ self.ind_fit ]      #len(ndim)
        self.fittruths = self.truths[ self.ind_fit ]    #len(ndim)
        self.theta_0 = buf['value'][ self.ind_fit ]     #len(ndim)
        
        if 'init_err' in buf.dtype.names:
            self.init_err = buf['init_err'][ self.ind_fit ] #len(ndim)
        else:
            self.init_err = 1e-8
        
        self.bounds = [ str(item).split(' ') for item in buf['bounds'][ self.ind_fit ] ] #len(ndim)
        for i, item in enumerate(self.bounds):
            if item[0] in ['uniform', 'normal']:
                self.bounds[i] = [ item[0], np.float(item[1]), np.float(item[2]) ]
            elif item[0] in ['trunc_normal']:
                self.bounds[i] = [ item[0], np.float(item[1]), np.float(item[2]), np.float(item[3]), np.float(item[4]) ]
            else:
                raise ValueError('Bounds have to be "uniform", "normal" or "trunc_normal". Input from "params.csv" was "'+self.bounds[i][0]+'".')
    
        self.ndim = len(self.theta_0)                   #len(ndim)

    
        #==========================================================================
        #::: luser proof: check if all initial guesses lie within their bounds
        #==========================================================================
        #TODO: make this part of the validate() function
        for th, b, key in zip(self.theta_0, self.bounds, self.fitkeys):
                  
            #:::: test bounds
            if (b[0] == 'uniform') and not (b[1] <= th <= b[2]): 
                raise ValueError('The initial guess for '+key+' lies outside of its bounds.')
                
            elif (b[0] == 'normal') and ( np.abs(th - b[1]) > 3*b[2] ):
                answer = input('The initial guess for '+key+' lies more than 3 sigma from its prior\n'+\
                      'What do you want to do?\n'+\
                      '1 : continue at any sacrifice \n'+\
                      '2 : stop and let me fix the params.csv file \n')
                if answer==1: 
                    pass
                else:
                    raise ValueError('User aborted the run.')
                    
            elif (b[0] == 'trunc_normal') and not (b[1] <= th <= b[2]): 
                raise ValueError('The initial guess for '+key+' lies outside of its bounds.')
                
            elif (b[0] == 'trunc_normal') and ( np.abs(th - b[3]) > 3*b[4] ): 
                answer = input('The initial guess for '+key+' lies more than 3 sigma from its prior\n'+\
                      'What do you want to do?\n'+\
                      '1 : continue at any sacrifice \n'+\
                      '2 : stop and let me fix the params.csv file \n')
                if answer==1: 
                    pass
                else:
                    raise ValueError('User aborted the run.')
            
            

    ###############################################################################
    #::: load data
    ###############################################################################
    def load_data(self):
        '''
        Example: 
        -------
            A lightcurve is stored as
                data['TESS']['time'], data['TESS']['flux'], etc.
            A RV curve is stored as
                data['HARPS']['time'], data['HARPS']['flux'], etc.
        '''
        self.fulldata = {}
        self.data = {}
        
        #======================================================================
        #::: photometry
        #======================================================================
        for inst in self.settings['inst_phot']:
            try:
                time, flux, flux_err, custom_series = np.genfromtxt(os.path.join(self.datadir,inst+'.csv'), delimiter=',', dtype=float, unpack=True)[0:4]     
            except:
                time, flux, flux_err = np.genfromtxt(os.path.join(self.datadir,inst+'.csv'), delimiter=',', dtype=float, unpack=True)[0:3]     
                custom_series = np.zeros_like(time)
            if any(np.isnan(time*flux*flux_err*custom_series)):
                raise ValueError('There are NaN values in "'+inst+'.csv". Please make sure everything is fine with your data, then exclude these rows from the file and restart.')
            if any(flux_err==0):
                raise ValueError('There are uncertainties with values of 0 in "'+inst+'.csv". Please make sure everything is fine with your data, then exclude these rows from the file and restart.')
            if any(flux_err<0):
                raise ValueError('There are uncertainties with negative values in "'+inst+'.csv". Please make sure everything is fine with your data, then exclude these rows from the file and restart.')
            if not all(np.diff(time)>=0):
                raise ValueError('The time array in "'+inst+'.csv" is not sorted. Please make sure the file is not corrupted, then sort it by time and restart.')
            elif not all(np.diff(time)>0):
                warnings.warn('There are repeated time stamps in the time array in "'+inst+'.csv". Please make sure the file is not corrupted (e.g. insuffiecient precision in your time stamps).')
#                overwrite = str(input('There are repeated time stamps in the time array in "'+inst+'.csv". Please make sure the file is not corrupted (e.g. insuffiecient precision in your time stamps).'+\
#                                      'What do you want to do?\n'+\
#                                      '1 : continue and hope for the best; no risk, no fun; #yolo\n'+\
#                                      '2 : abort\n'))
#                if (overwrite == '1'):
#                    pass
#                else:
#                    raise ValueError('User aborted operation.')
                
            self.fulldata[inst] = {
                          'time':time,
                          'flux':flux,
                          'err_scales_flux':flux_err/np.nanmean(flux_err),
                          'custom_series':custom_series
                         }
            if (self.settings['fast_fit']) and (len(self.settings['inst_phot'])>0): 
                time, flux, flux_err, custom_series = self.reduce_phot_data(time, flux, flux_err, custom_series=custom_series, inst=inst)
            self.data[inst] = {
                          'time':time,
                          'flux':flux,
                          'err_scales_flux':flux_err/np.nanmean(flux_err),
                          'custom_series':custom_series
                         }
            
        #======================================================================
        #::: RV
        #======================================================================
        for inst in self.settings['inst_rv']:
            try:
                time, rv, rv_err, custom_series = np.genfromtxt( os.path.join(self.datadir,inst+'.csv'), delimiter=',', dtype=float, unpack=True)[0:4]       
            except:
                time, rv, rv_err = np.genfromtxt( os.path.join(self.datadir,inst+'.csv'), delimiter=',', dtype=float, unpack=True)[0:3]              
                custom_series = np.zeros_like(time)
            if any(np.isnan(time*rv*rv_err*custom_series)):
                raise ValueError('There are NaN values in "'+inst+'.csv". Please make sure everything is fine with your data, then exclude these rows from the file and restart.')
            #aCkTuaLLLyy rv_err=0 is ok, since we add a jitter term here anyway (instead of scaling)
            # if any(rv_err==0):
            #     raise ValueError('There are uncertainties with values of 0 in "'+inst+'.csv". Please make sure everything is fine with your data, then exclude these rows from the file and restart.')
            if any(rv_err<0):
                raise ValueError('There are uncertainties with negative values in "'+inst+'.csv". Please make sure everything is fine with your data, then exclude these rows from the file and restart.')
            if not all(np.diff(time)>0):
                raise ValueError('Your time array in "'+inst+'.csv" is not sorted. You will want to check that...')
            self.data[inst] = {
                          'time':time,
                          'rv':rv,
                          'white_noise_rv':rv_err,
                          'custom_series':custom_series
                         }
            
        #======================================================================
        #::: RV2 (for detached binaries)
        #======================================================================
        for inst in self.settings['inst_rv2']:
            try:
                time, rv, rv_err, custom_series = np.genfromtxt( os.path.join(self.datadir,inst+'.csv'), delimiter=',', dtype=float, unpack=True)[0:4]       
            except:
                time, rv, rv_err = np.genfromtxt( os.path.join(self.datadir,inst+'.csv'), delimiter=',', dtype=float, unpack=True)[0:3]              
                custom_series = np.zeros_like(time)
            if not all(np.diff(time)>0):
                raise ValueError('Your time array in "'+inst+'.csv" is not sorted. You will want to check that...')
            self.data[inst] = {
                          'time':time,
                          'rv2':rv,
                          'white_noise_rv2':rv_err,
                          'custom_series':custom_series
                         }
        
        #======================================================================
        #::: also save the combined time series
        #::: for cases where all instruments are treated together
        #::: e.g. for stellar variability GPs
        #======================================================================
        self.data['inst_phot'] = {'time':[],'flux':[],'flux_err':[],'inst':[]}
        for inst in self.settings['inst_phot']:
            self.data['inst_phot']['time'] += list(self.data[inst]['time'])
            self.data['inst_phot']['flux'] += list(self.data[inst]['flux'])
            self.data['inst_phot']['flux_err'] += [inst]*len(self.data[inst]['time']) #errors will be sampled/derived later
            self.data['inst_phot']['inst'] += [inst]*len(self.data[inst]['time'])
        ind_sort = np.argsort(self.data['inst_phot']['time'])
        self.data['inst_phot']['ind_sort'] = ind_sort
        self.data['inst_phot']['time'] = np.array(self.data['inst_phot']['time'])[ind_sort]
        self.data['inst_phot']['flux'] = np.array(self.data['inst_phot']['flux'])[ind_sort]
        self.data['inst_phot']['flux_err'] = np.array(self.data['inst_phot']['flux_err'])[ind_sort]      
        self.data['inst_phot']['inst'] = np.array(self.data['inst_phot']['inst'])[ind_sort]
    
        self.data['inst_rv'] = {'time':[],'rv':[],'rv_err':[],'inst':[]}
        for inst in self.settings['inst_rv']:
            self.data['inst_rv']['time'] += list(self.data[inst]['time'])
            self.data['inst_rv']['rv'] += list(self.data[inst]['rv'])
            self.data['inst_rv']['rv_err'] += list(np.nan*self.data[inst]['rv']) #errors will be sampled/derived later
            self.data['inst_rv']['inst'] += [inst]*len(self.data[inst]['time'])
        ind_sort = np.argsort(self.data['inst_rv']['time'])
        self.data['inst_rv']['ind_sort'] = ind_sort
        self.data['inst_rv']['time'] = np.array(self.data['inst_rv']['time'])[ind_sort]
        self.data['inst_rv']['rv'] = np.array(self.data['inst_rv']['rv'])[ind_sort]
        self.data['inst_rv']['rv_err'] = np.array(self.data['inst_rv']['rv_err'])[ind_sort]   
        self.data['inst_rv']['inst'] = np.array(self.data['inst_rv']['inst'])[ind_sort]
    
        self.data['inst_rv2'] = {'time':[],'rv2':[],'rv2_err':[],'inst':[]}
        for inst in self.settings['inst_rv2']:
            self.data['inst_rv2']['time'] += list(self.data[inst]['time'])
            self.data['inst_rv2']['rv2'] += list(self.data[inst]['rv2'])
            self.data['inst_rv2']['rv2_err'] += list(np.nan*self.data[inst]['rv2']) #errors will be sampled/derived later
            self.data['inst_rv2']['inst'] += [inst]*len(self.data[inst]['time'])
        ind_sort = np.argsort(self.data['inst_rv2']['time'])
        self.data['inst_rv2']['ind_sort'] = ind_sort
        self.data['inst_rv2']['time'] = np.array(self.data['inst_rv2']['time'])[ind_sort]
        self.data['inst_rv2']['rv2'] = np.array(self.data['inst_rv2']['rv2'])[ind_sort]
        self.data['inst_rv2']['rv2_err'] = np.array(self.data['inst_rv2']['rv2_err'])[ind_sort]   
        self.data['inst_rv2']['inst'] = np.array(self.data['inst_rv2']['inst'])[ind_sort]

        
            
    ###############################################################################
    #::: change epoch
    ###############################################################################

    def my_truncnorm_isf(q,a,b,mean,std):
        a_scipy = 1.*(a - mean) / std
        b_scipy = 1.*(b - mean) / std
        return truncnorm.isf(q,a_scipy,b_scipy,loc=mean,scale=std)


    def change_epoch(self):
        '''
        change epoch entry from params.csv to set epoch into the middle of the range
        '''
        
        self.logprint('\nShifting epochs into the data center')
        self.logprint('------------------------------------')
        
        #::: for all companions
        for companion in self.settings['companions_all']:
            
            self.logprint('Companion',companion)
            self.logprint('\tinput epoch:',self.params[companion+'_epoch'])
            
            #::: get data time range
            alldata = []
            for inst in self.settings['inst_for_'+companion+'_epoch']:
                alldata += list(self.data[inst]['time'])
            start = np.nanmin( alldata )
            end = np.nanmax( alldata )
            
            #::: get the given values
            user_epoch  = 1.*self.params[companion+'_epoch']
            period      = 1.*self.params[companion+'_period']
#            buf = self.bounds[ind_e].copy()
                
            #::: calculate the true first_epoch
            if 'fast_fit_width' in self.settings and self.settings['fast_fit_width'] is not None:
                width = self.settings['fast_fit_width']
            else:
                width = 0
            first_epoch = get_first_epoch(alldata, self.params[companion+'_epoch'], self.params[companion+'_period'], width=width)
            
            #::: calculate the mid_epoch (in the middle of the data set)
            N = int(np.round((end-start)/2./period))
            self.settings['mid_epoch'] = first_epoch + N * period
            
            #::: calculate how much the user_epoch has to be shifted to get the mid_epoch
            N_shift = int(np.round((self.settings['mid_epoch']-user_epoch)/period))
            
            #::: set the new initial guess (and truth)
            self.params[companion+'_epoch'] = 1.*self.settings['mid_epoch']
           
            #::: also shift the truth (implies that the turth epoch is set where the initial guess is)
            try:
                ind_e = np.where(self.fitkeys==companion+'_epoch')[0][0]
                ind_p = np.where(self.fitkeys==companion+'_period')[0][0]
                N_truth_shift = int(np.round((self.settings['mid_epoch']-self.fittruths[ind_e])/self.fittruths[ind_p]))
                self.fittruths[ind_e] += N_truth_shift * self.fittruths[ind_p]
            except:
                pass
            
            #::: if a fit param, also update the bounds accordingly
            if (N_shift != 0) and (companion+'_epoch' in self.fitkeys):
                ind_e = np.where(self.fitkeys==companion+'_epoch')[0][0]
                ind_p = np.where(self.fitkeys==companion+'_period')[0][0]
                
#                print('\n')
#                print('############################################################################')
#                print('user_epoch', user_epoch, self.bounds[ind_e])
#                print('user_period', period, self.bounds[ind_p])
#                print('----------------------------------------------------------------------------')
                  
                #::: set the new initial guess
                self.theta_0[ind_e] = 1.*self.settings['mid_epoch']
                
                #::: get the bounds / errors
                #::: if the epoch and period priors are both uniform
                if (self.bounds[ind_e][0] == 'uniform') & (self.bounds[ind_p][0] == 'uniform'):
                    if N_shift > 0:
                        self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * self.bounds[ind_p][1] #lower bound
                        self.bounds[ind_e][2] = self.bounds[ind_e][2] + N_shift * self.bounds[ind_p][2] #upper bound
                    elif N_shift < 0:
                        self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * self.bounds[ind_p][2] #lower bound; period bounds switched if N_shift is negative
                        self.bounds[ind_e][2] = self.bounds[ind_e][2] + N_shift * self.bounds[ind_p][1] #upper bound; period bounds switched if N_shift is negative
                
                #::: if the epoch and period priors are both normal
                elif (self.bounds[ind_e][0] == 'normal') & (self.bounds[ind_p][0] == 'normal'):
                    self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * self.bounds[ind_p][1] #mean (in case the prior-mean is not the initial-guess-mean)
                    self.bounds[ind_e][2] = np.sqrt( self.bounds[ind_e][2]**2 + N_shift**2 * self.bounds[ind_p][2]**2 ) #std (in case the prior-mean is not the initial-guess-mean)
                                        
                #::: if the epoch and period priors are both trunc_normal
                elif (self.bounds[ind_e][0] == 'trunc_normal') & (self.bounds[ind_p][0] == 'trunc_normal'):
                    if N_shift > 0:
                        self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * self.bounds[ind_p][1] #lower bound
                        self.bounds[ind_e][2] = self.bounds[ind_e][2] + N_shift * self.bounds[ind_p][2] #upper bound
                    elif N_shift < 0:
                        self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * self.bounds[ind_p][2] #lower bound; period bounds switched if N_shift is negative
                        self.bounds[ind_e][2] = self.bounds[ind_e][2] + N_shift * self.bounds[ind_p][1] #upper bound; period bounds switched if N_shift is negative
                    self.bounds[ind_e][3] = self.bounds[ind_e][3] + N_shift * self.bounds[ind_p][3] #mean (in case the prior-mean is not the initial-guess-mean)
                    self.bounds[ind_e][4] = np.sqrt( self.bounds[ind_e][4]**2 + N_shift**2 * self.bounds[ind_p][4]**2 ) #std (in case the prior-mean is not the initial-guess-mean)
            
                #::: if the epoch prior is uniform and period prior is normal
                elif (self.bounds[ind_e][0] == 'uniform') & (self.bounds[ind_p][0] == 'normal'):
                    self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * (period + self.bounds[ind_p][2]) #lower bound epoch + Nshift * period + Nshift * std_period
                    self.bounds[ind_e][2] = self.bounds[ind_e][2] + N_shift * (period + self.bounds[ind_p][2]) #upper bound + Nshift * period + Nshift * std_period

                #::: if the epoch prior is uniform and period prior is trunc_normal
                elif (self.bounds[ind_e][0] == 'uniform') & (self.bounds[ind_p][0] == 'trunc_normal'):
                    self.bounds[ind_e][1] = self.bounds[ind_e][1] + N_shift * (period + self.bounds[ind_p][4]) #lower bound epoch + Nshift * period + Nshift * std_period
                    self.bounds[ind_e][2] = self.bounds[ind_e][2] + N_shift * (period + self.bounds[ind_p][4]) #upper bound + Nshift * period + Nshift * std_period

                elif (self.bounds[ind_e][0] == 'normal') & (self.bounds[ind_p][0] == 'uniform'):
                    raise ValueError('shift_epoch with different priors for epoch and period is not yet implemented.')
                    
                elif (self.bounds[ind_e][0] == 'normal') & (self.bounds[ind_p][0] == 'trunc_normal'):
                    raise ValueError('shift_epoch with different priors for epoch and period is not yet implemented.')
                    
                elif (self.bounds[ind_e][0] == 'trunc_normal') & (self.bounds[ind_p][0] == 'uniform'):
                    raise ValueError('shift_epoch with different priors for epoch and period is not yet implemented.')
                    
                elif (self.bounds[ind_e][0] == 'trunc_normal') & (self.bounds[ind_p][0] == 'normal'):
                    raise ValueError('shift_epoch with different priors for epoch and period is not yet implemented.')
                    
                else:
                    raise ValueError('Parameters "bounds" have to be "uniform", "normal" or "trunc_normal".')
                    
        
                self.logprint('\tshifted epoch:',self.params[companion+'_epoch'])
                self.logprint('\tshifted by',N_shift,'periods')
                


    ###############################################################################
    #::: reduce_phot_data
    ###############################################################################
    def reduce_phot_data(self, time, flux, flux_err, custom_series=None, inst=None):
        ind_in = []
              
        for companion in self.settings['companions_phot']:
            epoch  = self.params[companion+'_epoch']
            period = self.params[companion+'_period']
            width  = self.settings['fast_fit_width']
            if self.settings['secondary_eclipse']:
                ind_ecl1x, ind_ecl2x, ind_outx = index_eclipses(time,epoch,period,width,width) #TODO: currently this assumes width_occ == width_tra
                ind_in += list(ind_ecl1x)
                ind_in += list(ind_ecl2x)
                self.fulldata[inst][companion+'_ind_ecl1'] = ind_ecl1x
                self.fulldata[inst][companion+'_ind_ecl2'] = ind_ecl2x
                self.fulldata[inst][companion+'_ind_out'] = ind_outx
            else:
                ind_inx, ind_outx = index_transits(time,epoch,period,width)
                ind_in += list(ind_inx)
                self.fulldata[inst][companion+'_ind_in'] = ind_inx
                self.fulldata[inst][companion+'_ind_out'] = ind_outx
                
        ind_in = np.sort(np.unique(ind_in))
        self.fulldata[inst]['all_ind_in'] = ind_in
        self.fulldata[inst]['all_ind_out'] = np.delete( np.arange(len(self.fulldata[inst]['time'])), ind_in )
        
        if len(ind_in)==0:
            raise ValueError(inst+'.csv does not contain any in-transit data. Check that your epoch and period guess are correct.')
        
        time = time[ind_in]
        flux = flux[ind_in]
        flux_err = flux_err[ind_in]
        if custom_series is None: 
            return time, flux, flux_err
        else:
            custom_series = custom_series[ind_in]
            return time, flux, flux_err, custom_series
    
    
    
    ###############################################################################
    #::: setup TTV fit (if chosen)
    ###############################################################################
    def setup_ttv_fit(self):
        '''
        this must be run *after* reduce_phot_data()
        '''
        
        #::: the window we choose to look for transits is determined by fast_fit_width
        window = self.settings['fast_fit_width']
        
        #::: for each companion, stitch together all the time stamps observed by all photometric instruments
        #::: and check which of these times overlap with a potential transit window (determined by fast_fit_width)
        for companion in self.settings['companions_phot']:
            times_combined = []
            for inst in self.settings['inst_phot']:
                times_combined += list(self.data[inst]['time'])
            times_combined = np.sort(times_combined)
            
            self.data[companion+'_tmid_observed_transits'] = get_tmid_observed_transits(times_combined,
                                                                                        self.params[companion+'_epoch'],
                                                                                        self.params[companion+'_period'],
                                                                                        window)
            
            for inst in self.settings['inst_phot']:
                time = self.data[inst]['time']
                for i, t in enumerate(self.data[companion+'_tmid_observed_transits']):
                    ind = np.where((time >= (t - window/2.)) & (time <= (t + window/2.)))[0]
                    self.data[inst][companion+'_ind_time_transit_'+str(i+1)] = ind
                    self.data[inst][companion+'_time_transit_'+str(i+1)] = time[ind]
                    

            #::: THE FOLLOWING PART MOVED INTO THE SEPARATE SCRIPT "PREPARE_TTV_FIT.PY"
            #::: plots
            # if self.settings['fit_ttvs']:  
            #     flux_min = np.nanmin(all_flux)
            #     flux_max = np.nanmax(all_flux)
            #     N_days = int( np.max(all_times) - np.min(all_times) )
            #     figsizex = np.min( [1, int(N_days/20.)] )*5
            #     fig, ax = plt.subplots(figsize=(figsizex, 4)) #figsize * 5 for every 20 days
            #     for inst in self.settings['inst_phot']:
            #         ax.plot(self.data[inst]['time'], self.data[inst]['flux'],ls='none',marker='.',label=inst)
            #     ax.plot( self.data[companion+'_tmid_observed_transits'], np.ones_like(self.data[companion+'_tmid_observed_transits'])*0.995*flux_min, 'k^' )
            #     for i, tmid in enumerate(self.data[companion+'_tmid_observed_transits']):
            #         ax.text( tmid, 0.9925*flux_min, str(i+1), ha='center' )  
            #     ax.set(ylim=[0.99*flux_min, flux_max], xlabel='Time (BJD)', ylabel='Realtive Flux') 
            #     if not os.path.exists( os.path.join(self.datadir,'results') ):
            #         os.makedirs(os.path.join(self.datadir,'results'))
            #     ax.legend()
            #     fname = os.path.join(self.datadir,'results','preparation_for_TTV_fit_'+companion+'.pdf')
            #     if os.path.exists(fname):
            #         overwrite = str(input('Figure "preparation_for_TTV_fit_'+companion+'.pdf" already exists.\n'+\
            #                               'What do you want to do?\n'+\
            #                               '1 : overwrite it\n'+\
            #                               '2 : skip it and move on\n'))
            #         if (overwrite == '1'):
            #             fig.savefig(fname, bbox_inches='tight' )    
            #         else:
            #             pass        
            #     plt.close(fig)
            
                
                
    ###############################################################################
    #::: stellar priors
    ###############################################################################
    def load_stellar_priors(self, N_samples=10000):
        if os.path.exists(os.path.join(self.datadir,'params_star.csv')) and (self.settings['use_host_density_prior'] is True):
            buf = np.genfromtxt( os.path.join(self.datadir,'params_star.csv'), delimiter=',', names=True, dtype=None, encoding='utf-8', comments='#' )
            radius = simulate_PDF(buf['R_star'], buf['R_star_lerr'], buf['R_star_uerr'], size=N_samples, plot=False) * 6.957e10 #in cgs
            mass = simulate_PDF(buf['M_star'], buf['M_star_lerr'], buf['M_star_uerr'], size=N_samples, plot=False) * 1.9884754153381438e+33 #in cgs
            volume = (4./3.)*np.pi*radius**3 #in cgs
            density = mass / volume #in cgs
            self.params_star = {'R_star_median':buf['R_star'],
                                'R_star_lerr':buf['R_star_lerr'],
                                'R_star_uerr':buf['R_star_uerr'],
                                'M_star_median':buf['M_star'],
                                'M_star_lerr':buf['M_star_lerr'],
                                'M_star_uerr':buf['M_star_uerr']
                                }
            self.external_priors['host_density'] = ['normal', np.median(density), np.max( [np.median(density)-np.percentile(density,16), np.percentile(density,84)-np.median(density)] ) ] #in cgs
            
            


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  2 10:54:58 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

#::: modules
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize, basinhopping
from scipy.stats import skewnorm

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: random seed
np.random.seed(42)




def simulate_PDF(median, lower_err, upper_err, size=1, plot=True):
    '''
    Simulates a draw of posterior samples from a value and asymmetric errorbars
    by assuming the underlying distribution is a skewed normal distribution.
    
    Developed to estimate PDFs from literature exoplanet parameters that did not report their MCMC chains.
    
    Inputs:
    -------
    median : float
        the median value that was reported
    lower_err : float
        the lower errorbar that was reported
    upper_err : float
        the upper errorbar that was reported
    size : int
        the number of samples to be drawn
        
    Returns:
    --------
    samples : array of float
        the samples drawn from the simulated skrewed normal distribution
    '''
    
    alpha, loc, scale = calculate_skewed_normal_params(median, lower_err, upper_err)
    samples = skewnorm.rvs(alpha, loc=loc, scale=scale, size=size)
    
    if plot==False:
        return samples

    else:
        lower_err = np.abs(lower_err)
        upper_err = np.abs(upper_err)
        x = np.linspace(np.min(samples)-0.1*np.ptp(samples), np.max(samples)+0.1*np.ptp(samples), 1000)
        # x = np.arange(median-4*lower_err, median+4*upper_err, 0.001)
        fig = plt.figure()
        for i in range(3): plt.axvline([median-lower_err, median, median+upper_err][i], color='k', lw=2)
        plt.plot( x, skewnorm.pdf(x, alpha, loc=loc, scale=scale), 'r-', lw=2 )
        fit_percentiles = skewnorm.ppf([0.15865, 0.5, 0.84135], alpha, loc=loc, scale=scale)
        for i in range(3): plt.axvline(fit_percentiles[i], color='r', ls='--', lw=2)
        plt.hist(samples, density=True, color='r', alpha=0.5, bins=100)
        print('\n')
        print('Input', median-lower_err, median, median+upper_err)
        print('Output', fit_percentiles)
        plt.text(0.95, 0.95, 'Black: Input\nRed: Output', ha='right', va='top', transform=plt.gca().transAxes)
        return samples, fig



def calculate_skewed_normal_params(median, lower_err, upper_err):
    '''
    Fits a screwed normal distribution via its CDF to the [16,50,84]-percentiles
    
    Inputs:
    -------
    median : float
        the median value that was reported
    lower_err : float
        the lower errorbar that was reported
    upper_err : float
        the upper errorbar that was reported
    size : int
        the number of samples to be drawn
        
    Returns:
    --------
    alpha : float
        the skewness parameter
    loc : float
        the mean of the fitted skewed normal distribution
    scale : float
        the std of the fitted skewed normal distribution
    '''
    
    lower_err = np.abs(lower_err)
    upper_err = np.abs(upper_err)
    reference = np.array([(median-lower_err), median, (median+upper_err)])
    
    
    def fake_lnlike(p):
        alpha, loc, scale = p
        
        #::: way 1: easier to read, but slower; fake log likelihood
        # eq1 = skewnorm.ppf(0.5, alpha, loc=loc, scale=scale) - median 
        # eq2 = skewnorm.ppf(0.15865, alpha, loc=loc, scale=scale) - (median-lower_err)
        # eq3 = skewnorm.ppf(0.84135, alpha, loc=loc, scale=scale) - (median+upper_err)
        # fake_lnlike = 0.5 * np.log( eq1**2 + eq2**2 + eq3**2 ) #fake log likelihod
        
        #::: way 2: pythonic; simple chi squared
        ppf = skewnorm.ppf([0.15865, 0.5, 0.84135], alpha, loc=loc, scale=scale) #np array
        fake_lnlike = np.sum( (ppf - reference)**2 ) #simple chi squared
        
        #::: way 2: pythonic; fake log likehood, just 'cause we're feeling fancy
        # ppf = skewnorm.ppf([0.15865, 0.5, 0.84135], alpha, loc=loc, scale=scale) #np array
        # fake_lnlike = 0.5 * np.log( np.sum( (ppf - reference)**2 ) ) #fake log likelihod
        
        if np.isnan(fake_lnlike): 
            return np.inf
        else:
            return fake_lnlike


    #TODO: 
    #scipy minimize is really bad because it depends so strongly on the initial guess
    #and likes to get stuck in a local minima, which is the worst possible outcome
    #for our purpose here. We only use it because it is fast. Think about replacing
    #it with something more robust in the future though, maybe a short MCMC chain or the like.
    #The alpha_guess hack below seems to get around its weakness in finding the right alpha for now.

        
    #::: initial guess for loc and scale
    loc_guess = median
    scale_guess = np.mean([lower_err, upper_err])
    # print('\n')

    
    #::: way 1: choose alpha_guess depending on the errors and hope for the best
    # if lower_err == upper_err:
    #     alpha_guess = 0
    # elif lower_err < upper_err:
    #     alpha_guess = 1
    # elif lower_err > upper_err:
    #     alpha_guess = -1
    # initial_guess = (median, sigma_guess, alpha_guess) #sigma, omega, alpha 
    # sol = minimize(fake_lnlike, initial_guess, bounds=[(None,None), (0,None), (None,None)]) 
    # sigma, omega, alpha = sol.x
    
    
    #::: way 2: choose a few different alpha_guesses and compare (similar to basinhopping optimization)
    # initial_guess1 = None #just for printing
    sol = None
    for alpha_guess in [-10,-1,0,1,10]:
        initial_guess1 = (alpha_guess, loc_guess, scale_guess)
        sol1 = minimize(fake_lnlike, initial_guess1, bounds=[(None,None), (None,None), (0,None)]) 
        # print('sol1.fun', sol1.fun)
        if (sol is None) or (sol1.fun < sol.fun):
            # initial_guess = initial_guess1 #just for printing
            sol = sol1
            
            
    # print('best initial_guess:', initial_guess)
    # print('best solution:', sol)
    
    
    alpha, loc, scale = sol.x
    return alpha, loc, scale

    
    
if __name__ == '__main__':
    '''
    For testing, simulate a skewed normal distribution with parameters
    (alpha_0, loc_0, scale_0), perform the fit, and compare the resulting PDF.
    '''
    
    ###########################################################################
    # ::: three test cases
    ###########################################################################
    # simulate_PDF(1., 0.15, 0.15, size=10000) #R_star
    
    # simulate_PDF(1.895, 0.077, 0.1, size=10000) #M_star 1
    
    # simulate_PDF(1.895, 0.07, 0.15, size=10000) #M_star 2

    # simulate_PDF(84.3, -2.8, 1.3, size=10000) #inclination
    
    
    
    ###########################################################################
    # ::: simulate a skewed normal PDF, measure median and errors, and fit
    ###########################################################################
#    sigma_0 = 0.
#    omega_0 = 5.
#    alpha_0 = 2
#    
#    x = np.arange(-20+sigma_0,20+sigma_0,0.01)
#    y = skewnorm.pdf(x, alpha_0, loc=sigma_0, scale=omega_0)
#    plt.figure()
#    plt.plot( x, y, 'b-', lw=4 )
#    
#    median = skewnorm.ppf(0.5, alpha_0, loc=sigma_0, scale=omega_0)
#    lower_err = median - skewnorm.ppf(0.16, alpha_0, loc=sigma_0, scale=omega_0)
#    upper_err = skewnorm.ppf(0.84, alpha_0, loc=sigma_0, scale=omega_0) - median
#    plt.axvline(median)
#    plt.axvline(median-lower_err)
#    plt.axvline(median+upper_err)
#    
#    sigma, omega, alpha = calculate_skewed_normal_params(median, lower_err, upper_err)
#    plt.plot( x, skewnorm.pdf(x, alpha, loc=sigma, scale=omega), 'r--', lw=2 )



    ###########################################################################
    # ::: example inclination
    ###########################################################################
#    (median, lower_err, upper_err) = (84.3, -2.0, 1.3)
#    samples, fig = simulate_posterior_samples(median, lower_err, upper_err, size=1, plot=True)
#    print(np.percentile(samples, [16,50,84]))


#    lower_err = np.abs(lower_err)
#    upper_err = np.abs(upper_err)
#    
#    x = np.arange(median-4*lower_err, median+4*upper_err, 0.01)
#    plt.figure()
#    plt.axvline(median, color='k', lw=2)
#    plt.axvline(median-lower_err, color='k', lw=2)
#    plt.axvline(median+upper_err, color='k', lw=2)
#    
#    sigma, omega, alpha = calculate_skewed_normal_params(median, lower_err, upper_err)
#    plt.plot( x, skewnorm.pdf(x, alpha, loc=sigma, scale=omega), 'r-', lw=2 )
#    
#    fit_percentiles = skewnorm.ppf([0.16, 0.5, 0.84], alpha, loc=sigma, scale=omega)
#    for i in range(3): plt.axvline(fit_percentiles[i], color='r', ls='--', lw=2)
#    
#    rvs = simulate_posterior_samples(median, lower_err, upper_err, size=1000)
#    plt.hist(rvs, density=True, color='lightblue')
